from aiogram.fsm.storage.memory import MemoryStorage

from .base import BaseRoleProvider
from .protocol import IRoleProvider

__all__ = ("MemoryRoleProvider",)


class MemoryRoleProvider(BaseRoleProvider, IRoleProvider):
    """Simple in-memory role provider for testing and development."""

    def __init__(self, storage: MemoryStorage) -> None:
        """Initialize MemoryRoleProvider."""
        super().__init__(storage=storage)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from aiogram import Bot
from aiogram.filters import Filter
from aiogram.types import TelegramObject, User
from typing_extensions import override

from .data import RoleData

if TYPE_CHECKING:
    from raito.core.raito import Raito

__all__ = ("RoleFilter",)

FLAG_NAME = "raito__roles"


class RoleFilter(Filter):
    """Filter for checking user roles.

    This filter is used to verify whether the user associated with a Telegram event
    has a specific role assigned, such as "admin", "moderator", etc.

    It also attaches role metadata to the handler's flags for use in command registration
    and visualization logic.
    """

    def __init__(
        self,
        slug: str,
        name: str,
        description: str,
        emoji: str,
    ) -> None:
        """Initialize the RoleFilter.

        :param slug: Unique identifier of the role (e.g., "developer")
        :param name: Display name of the role (e.g., "Administrator")
        :param description: Description of the role
        :param emoji: Emoji used to visually represent the role
        """
        self.data = RoleData(slug=slug, name=name, description=description, emoji=emoji)

    @classmethod
    def from_data(cls, data: RoleData) -> RoleFilter:
        """Create a RoleFilter from a RoleData instance.

        :param data: RoleData instance containing role metadata
        :return: A new RoleFilter instance
        """
        return RoleFilter(
            slug=data.slug,
            name=data.name,
            description=data.description,
            emoji=data.emoji,
        )

    @override
    def update_handler_flags(self, flags: dict[str, Any]) -> None:
        """Attach role metadata to handler flags.

        This allows external tools to collect and display role-related constraints.
        """
        roles = flags.setdefault("roles", [])
        roles.append(self.data)

    @override
    async def __call__(
        self,
        event: TelegramObject,
        raito: Raito,
        bot: Bot,
        *args: Any,
        **kwargs: Any,
    ) -> bool:
        """Check if the user in the event has the required role.

        This method is automatically called by aiogram when the filter is used.
        It extracts the user from the event and checks the role manager.

        :param event: Telegram update object (e.g., Message, CallbackQuery)
        :param raito: Raito context object
        :return: Whether the user has the specified role
        :raises RuntimeError: If user could not be resolved from the event
        """
        user = getattr(event, "from_user", None)
        if not isinstance(user, User):
            msg = "Cannot resolve user from TelegramObject"
            raise RuntimeError(msg)

        return await raito.role_manager.has_role(bot.id, user.id, self.data.slug)



from collections.abc import AsyncGenerator, Callable
from contextlib import AbstractAsyncContextManager, asynccontextmanager, suppress
from typing import TypeAlias

from aiogram import Bot, Router

from raito.utils.helpers.safe_partial import safe_partial

__all__ = ("lifespan",)

FuncType: TypeAlias = Callable[..., AsyncGenerator[None, None]]
AsyncCtx: TypeAlias = AbstractAsyncContextManager[None]
LifespanStacks: TypeAlias = dict[int, list[AsyncCtx]]

_LIFESPAN_STACKS = "__lifespan_stacks__"


def _get_stack(router: Router) -> LifespanStacks:
    stacks = getattr(router, _LIFESPAN_STACKS, None)
    if stacks is None:
        stacks = {}
        setattr(router, _LIFESPAN_STACKS, stacks)
    return stacks


def lifespan(router: Router) -> Callable[[FuncType], FuncType]:
    """
    Register a lifespan function for a given router, similar to FastAPI's lifespan handler.
    The function must be an async generator: it runs setup before `yield`, and cleanup after.
    """

    def decorator(func: FuncType) -> FuncType:
        @asynccontextmanager
        async def context(**kwargs: dict[str, object]) -> AsyncGenerator[None, None]:
            gen = safe_partial(func, **kwargs)()
            await gen.__anext__()
            try:
                yield
            finally:
                with suppress(StopAsyncIteration):
                    await gen.__anext__()

        async def on_startup(**kwargs: dict[str, object]) -> None:
            bot = kwargs.get("bot")
            assert isinstance(bot, Bot), "Missing or invalid 'bot' in lifespan context"

            ctx = context(**kwargs)
            await ctx.__aenter__()
            _get_stack(router).setdefault(bot.id, []).append(ctx)

        async def on_shutdown(**kwargs: dict[str, object]) -> None:
            bot = kwargs.get("bot")
            assert isinstance(bot, Bot), "Missing or invalid 'bot' in lifespan context"

            stack = _get_stack(router).get(bot.id, [])
            for ctx in reversed(stack):
                await ctx.__aexit__(None, None, None)
            stack.clear()

        router.startup.register(on_startup)
        router.shutdown.register(on_shutdown)

        return func

    return decorator



from __future__ import annotations

from typing import TYPE_CHECKING

from aiogram import Router, html

from raito.plugins.commands import description, hidden
from raito.plugins.roles.roles import ADMINISTRATOR, DEVELOPER, OWNER
from raito.utils.ascii import AsciiTree, TreeNode
from raito.utils.filters import RaitoCommand

if TYPE_CHECKING:
    from aiogram import Bot
    from aiogram.types import Message

    from raito.core.raito import Raito

router = Router(name="raito.roles.staff")


@router.message(RaitoCommand("staff"), DEVELOPER | OWNER | ADMINISTRATOR)
@description("Shows users with roles")
@hidden
async def list_staff(message: Message, raito: Raito, bot: Bot) -> None:
    root = TreeNode("staff", is_folder=True)

    for role in raito.role_manager.available_roles:
        role_node = root.add_child(role.label, prefix=role.emoji, is_folder=True)
        user_ids = await raito.role_manager.get_users(bot_id=bot.id, role=role.slug)

        for user_id in user_ids:
            role_node.add_child(html.code(str(user_id)))

    tree = AsciiTree(folder_icon="", sort=False).render(root)
    text = html.bold("Current staff:") + "\n\n" + tree
    await message.answer(text, parse_mode="HTML")



from raito.utils.storages.json import JSONStorage

from .base import BaseRoleProvider
from .protocol import IRoleProvider

__all__ = ("JSONRoleProvider",)


class JSONRoleProvider(BaseRoleProvider, IRoleProvider):
    """JSON-based role provider for testing and development."""

    def __init__(self, storage: JSONStorage) -> None:
        """Initialize JSONRoleProvider."""
        super().__init__(storage=storage)



from __future__ import annotations

from typing import TYPE_CHECKING

from aiogram import F, Router, html
from aiogram.filters.callback_data import CallbackData
from aiogram.fsm.state import State, StatesGroup
from aiogram.types import Message
from aiogram.utils.keyboard import InlineKeyboardBuilder

from raito.plugins.commands import description
from raito.plugins.commands.flags import hidden
from raito.plugins.keyboards import dynamic
from raito.plugins.roles.data import RoleData
from raito.plugins.roles.roles import ADMINISTRATOR, DEVELOPER, OWNER
from raito.utils.filters import RaitoCommand

if TYPE_CHECKING:
    from aiogram.fsm.context import FSMContext
    from aiogram.types import CallbackQuery

    from raito.core.raito import Raito

router = Router(name="raito.roles.assign")


class AssignRoleCallback(CallbackData, prefix="rt_assign_role"):  # type: ignore[call-arg]
    """Callback data for assigning roles."""

    role_slug: str


class AssignRoleGroup(StatesGroup):
    """State group for assigning roles."""

    user_id = State()


@dynamic(2)
def roles_list_markup(builder: InlineKeyboardBuilder, roles: list[RoleData]) -> None:
    for role in roles:
        builder.button(
            text=role.emoji + " " + role.name,
            callback_data=AssignRoleCallback(role_slug=role.slug),
        )


@router.message(RaitoCommand("roles", "assign"), DEVELOPER | OWNER | ADMINISTRATOR)
@description("Assigns a role to a user")
@hidden
async def show_roles(message: Message, raito: Raito) -> None:
    await message.answer(
        "🎭 Select role to assign:",
        reply_markup=roles_list_markup(raito.role_manager.available_roles),
    )


@router.callback_query(AssignRoleCallback.filter(), DEVELOPER | OWNER | ADMINISTRATOR)
async def store_role(
    query: CallbackQuery,
    state: FSMContext,
    callback_data: AssignRoleCallback,
    raito: Raito,
) -> None:
    if not query.bot:
        await query.answer("🚫 Bot not found", show_alert=True)
        return
    if not isinstance(query.message, Message):
        await query.answer("🚫 Invalid message", show_alert=True)
        return

    role = raito.role_manager.get_role_data(callback_data.role_slug)
    await state.update_data(rt_selected_role=role.slug)
    await state.set_state(AssignRoleGroup.user_id)

    chat_id = query.message.chat.id
    await query.bot.send_message(
        chat_id=chat_id,
        text=f"{html.bold(role.label)}\n\n{html.blockquote(role.description)}",
        parse_mode="HTML",
    )
    await query.bot.send_message(chat_id=chat_id, text="👤 Enter user ID:")


@router.message(
    AssignRoleGroup.user_id,
    F.text and F.text.isdigit(),
    DEVELOPER | OWNER | ADMINISTRATOR,
)
async def assign_role(message: Message, raito: Raito, state: FSMContext) -> None:
    data = await state.get_data()
    role_slug = data.get("rt_selected_role")
    if role_slug is None:
        await message.answer("🚫 Role not selected")
        return
    if not message.from_user:
        await message.answer("🚫 User not found")
        return
    if not message.text or not message.text.isdigit():
        await message.answer("🚫 Invalid user ID")
        return
    if not message.bot:
        await message.answer("🚫 Bot instance not found")
        return

    await state.update_data(rt_selected_role=None)
    await state.set_state()

    role = raito.role_manager.get_role_data(role_slug)
    try:
        await raito.role_manager.assign_role(
            message.bot.id,
            message.from_user.id,
            int(message.text),
            role.slug,
        )
    except PermissionError:
        await message.answer("🚫 Permission denied")
        return

    await message.answer(f"❇️ User assigned to {html.bold(role.label)}", parse_mode="HTML")



from __future__ import annotations

from typing import TYPE_CHECKING

from aiogram.client.default import Default
from aiogram.types import InlineKeyboardMarkup
from aiogram.utils.keyboard import InlineKeyboardBuilder

from raito.plugins.pagination.enums import PaginationMode

from .base import BasePaginator

if TYPE_CHECKING:
    from aiogram.types import (
        InlineKeyboardButton,
        LinkPreviewOptions,
        Message,
        MessageEntity,
        ReplyParameters,
    )


__all__ = ("InlinePaginator",)


class InlinePaginator(BasePaginator):
    """Inline keyboard paginator."""

    def _validate_parameters(
        self,
        name: str,
        current_page: int,
        total_pages: int | None,
        limit: int,
    ) -> None:
        """Validate paginator parameters.

        :param name: pagination name
        :type name: str
        :param current_page: current page number
        :type current_page: int
        :param total_pages: total pages count
        :type total_pages: int | None
        :param limit: items per page
        :type limit: int
        :raises ValueError: if parameters are invalid
        """
        if limit > 90:
            raise ValueError("limit must be less than or equal to 90")

        return super()._validate_parameters(
            name=name,
            current_page=current_page,
            total_pages=total_pages,
            limit=limit,
        )

    @property
    def mode(self) -> PaginationMode:
        """Get inline pagination mode.

        :return: pagination mode
        :rtype: PaginationMode
        """
        return PaginationMode.INLINE

    def _get_content_builder(
        self,
        buttons: list[InlineKeyboardButton] | InlineKeyboardMarkup | None = None,
    ) -> InlineKeyboardBuilder:
        """Build content keyboard from buttons.

        :param buttons: content buttons or markup
        :type buttons: list[InlineKeyboardButton] | InlineKeyboardMarkup | None
        :return: keyboard builder with content
        :rtype: InlineKeyboardBuilder
        """
        builder = InlineKeyboardBuilder()

        if isinstance(buttons, list):
            builder.row(*buttons, width=1)
        elif isinstance(buttons, InlineKeyboardMarkup):
            content_builder = InlineKeyboardBuilder.from_markup(buttons)
            builder.attach(content_builder)

        return builder

    def _get_reply_markup(
        self,
        reply_markup: InlineKeyboardMarkup | None = None,
        buttons: list[InlineKeyboardButton] | InlineKeyboardMarkup | None = None,
    ) -> InlineKeyboardMarkup:
        """Build reply markup with content and navigation.

        :param buttons: content buttons or markup
        :type buttons: list[InlineKeyboardButton] | InlineKeyboardMarkup | None
        :return: complete keyboard markup
        :rtype: InlineKeyboardMarkup
        """
        builder = InlineKeyboardBuilder()

        if isinstance(buttons, list):
            builder.row(*buttons, width=1)
        elif isinstance(buttons, InlineKeyboardMarkup):
            builder.attach(builder.from_markup(buttons))

        if reply_markup is not None:
            builder.attach(builder.from_markup(reply_markup))
        else:
            builder.attach(builder.from_markup(self.build_navigation()))

        return builder.as_markup()

    async def answer(
        self,
        text: str,
        buttons: list[InlineKeyboardButton] | InlineKeyboardMarkup | None = None,
        parse_mode: str | Default | None = None,
        entities: list[MessageEntity] | None = None,
        link_preview_options: LinkPreviewOptions | Default | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | Default | None = None,
        allow_paid_broadcast: bool | None = None,
        message_effect_id: str | None = None,
        reply_parameters: ReplyParameters | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
        allow_sending_without_reply: bool | None = None,
        disable_web_page_preview: bool | Default | None = None,
        reply_to_message_id: int | None = None,
    ) -> Message:
        """Send or edit paginated message.

        :param text: message text
        :type text: str
        :param buttons: content buttons
        :type buttons: list[list[InlineKeyboardButton]]
        :param parse_mode: text parse mode
        :type parse_mode: str | Default | None
        :param entities: message entities
        :type entities: list[MessageEntity] | None
        :param link_preview_options: link preview settings
        :type link_preview_options: LinkPreviewOptions | Default | None
        :param disable_notification: disable notification
        :type disable_notification: bool | None
        :param protect_content: protect content
        :type protect_content: bool | Default | None
        :param allow_paid_broadcast: allow paid broadcast
        :type allow_paid_broadcast: bool | None
        :param message_effect_id: message effect id
        :type message_effect_id: int | None
        :param reply_parameters: reply parameters
        :type reply_parameters: ReplyParameters | None
        :param reply_markup: custom reply markup
        :type reply_markup: InlineKeyboardMarkup | None
        :param allow_sending_without_reply: allow sending without reply
        :type allow_sending_without_reply: bool | None
        :param disable_web_page_preview: disable web page preview
        :type disable_web_page_preview: bool | Default | None
        :param reply_to_message_id: reply to message id
        :type reply_to_message_id: int | None
        :return: paginated message
        :rtype: Message
        :raises RuntimeError: if bot instance not set
        """
        if not self.bot:
            raise RuntimeError("Bot not set via PaginatorMiddleware")

        parse_mode = parse_mode or Default("parse_mode")
        link_preview_options = link_preview_options or Default("link_preview")
        protect_content = protect_content or Default("protect_content")
        disable_web_page_preview = disable_web_page_preview or Default("link_preview_is_disabled")

        reply_markup = (
            reply_markup
            if reply_markup is not None and buttons is None
            else self._get_reply_markup(reply_markup, buttons)
        )

        if self.existing_message is None:
            self.existing_message = await self.bot.send_message(
                chat_id=self.chat_id,
                text=text,
                parse_mode=parse_mode,
                entities=entities,
                link_preview_options=link_preview_options,
                disable_notification=disable_notification,
                protect_content=protect_content,
                allow_paid_broadcast=allow_paid_broadcast,
                message_effect_id=message_effect_id,
                reply_parameters=reply_parameters,
                reply_markup=reply_markup,
                allow_sending_without_reply=allow_sending_without_reply,
                disable_web_page_preview=disable_web_page_preview,
                reply_to_message_id=reply_to_message_id,
            )
        elif text != self.existing_message.text:
            await self.existing_message.edit_text(text=text, reply_markup=reply_markup)
        else:
            await self.existing_message.edit_reply_markup(reply_markup=reply_markup)

        return self.existing_message



from typing import Protocol, runtime_checkable

__all__ = ("IRoleProvider",)


@runtime_checkable
class IRoleProvider(Protocol):
    """Protocol for providers that manage user roles."""

    async def get_role(self, bot_id: int, user_id: int) -> str | None:
        """Get the role for a specific user.

        :param bot_id: The Telegram bot ID
        :param user_id: The Telegram user ID
        :return: The role slug or None if not found
        """
        ...

    async def set_role(self, bot_id: int, user_id: int, role_slug: str) -> None:
        """Set the role for a specific user.

        :param bot_id: The Telegram bot ID
        :param user_id: The Telegram user ID
        :param role_slug: The role slug to assign
        """
        ...

    async def remove_role(self, bot_id: int, user_id: int) -> None:
        """Remove the role for a specific user.

        :param bot_id: The Telegram bot ID
        :param user_id: The Telegram user ID
        """
        ...

    async def migrate(self) -> None:
        """Initialize the storage backend (create tables, etc.)."""
        ...

    async def get_users(self, bot_id: int, role_slug: str) -> list[int]:
        """Get all users with a specific role.

        :param bot_id: The Telegram bot ID
        :param role_slug: The role slug to check for
        :return: A list of Telegram user IDs
        """
        ...



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from aiogram.filters import Filter, or_f

if TYPE_CHECKING:
    from .filter import RoleFilter

__all__ = (
    "RoleConstraint",
    "RoleGroupConstraint",
)


class RoleConstraint(Filter):
    """Wrapper around RoleFilter that supports logical OR composition."""

    def __init__(self, filter: RoleFilter) -> None:
        """Initialize RoleConstraint.

        :param filter: An instance of RoleFilter
        """
        self.filter = filter

    def __or__(self, other: RoleConstraint | RoleGroupConstraint) -> RoleGroupConstraint:
        """Combine this marker with another using the `|` operator.

        :param other: Another RoleConstraint or RoleGroupConstraint
        :return: A combined RoleGroupConstraint
        """
        if isinstance(other, RoleGroupConstraint):
            return RoleGroupConstraint(self, *other.filters)
        return RoleGroupConstraint(self, other)

    __ror__ = __or__

    async def __call__(self, *args: Any, **kwargs: Any) -> bool:  # noqa: ANN401
        """Delegate the call to the inner filter."""
        return await self.filter(*args, **kwargs)


class RoleGroupConstraint(Filter):
    """Logical group of multiple RoleConstraints combined via OR."""

    def __init__(self, *filters: RoleConstraint) -> None:
        """Initialize RoleGroupConstraint.

        :param filters: One or more RoleConstraint instances
        """
        self.filters = filters

    def __or__(self, other: RoleConstraint | RoleGroupConstraint) -> RoleGroupConstraint:
        """Extend the group with another marker or group.

        :param other: Another RoleConstraint or RoleGroupConstraint
        :return: New RoleGroupConstraint with all combined filters
        """
        if isinstance(other, RoleGroupConstraint):
            return RoleGroupConstraint(*self.filters, *other.filters)
        return RoleGroupConstraint(*self.filters, other)

    async def __call__(self, *args: Any, **kwargs: Any) -> bool:  # noqa: ANN401
        """Convert the group into an `_OrFilter` filter."""
        or_filter = or_f(*[f.filter for f in self.filters])
        value = await or_filter(*args, **kwargs)
        return bool(value)



from .base import BasePaginator
from .inline import InlinePaginator
from .photo import PhotoPaginator
from .protocol import IPaginator
from .text import TextPaginator

__all__ = (
    "BasePaginator",
    "IPaginator",
    "InlinePaginator",
    "PhotoPaginator",
    "TextPaginator",
)



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from aiogram.dispatcher.flags import Flag, FlagDecorator

if TYPE_CHECKING:
    from aiogram.utils.i18n.lazy_proxy import LazyProxy  # type: ignore

__all__ = ("description", "hidden", "params")


def description(description: str | LazyProxy) -> FlagDecorator:
    """Attach a description to the command handler for use in command registration (e.g., :code:`raito.register_commands`)

    The description will be shown in the Telegram command list (via :code:`set_my_commands`)
    Supports internationalization via `LazyProxy`.

    :param description: A string or LazyProxy representing the description.
    :return: A FlagDecorator to be applied to the handler.
    """
    return FlagDecorator(Flag("raito__description", value=description))


def hidden(func: Callable) -> Callable[..., Any]:
    """Mark a command handler as hidden from the command list.

    Hidden handlers will not be included in Telegram's slash commands when calling :code:`raito.register_commands`.

    :param func: The command handler to mark as hidden.
    :return: The wrapped handler.
    """
    return FlagDecorator(Flag("raito__hidden", value=True))(func)


def params(**kwargs: type[str] | type[int] | type[bool] | type[float]) -> FlagDecorator:
    """Define expected parameters and their types for command parsing.

    This acts as a lightweight argument extractor and validator for commands.
    For example, :code:`@rt.params(user_id=int)` will extract :code:`user_id=1234` from a command like :code:`/ban 1234`.

    Example:

        .. code-block:: python

            @router.message(filters.Command("ban"))
            @rt.params(user_id=int)
            def ban(message: Message, user_id: int):
                ...

    :param kwargs: A mapping of parameter names to their expected types.
    :return: A FlagDecorator to be applied to the handler with param data.
    """
    return FlagDecorator(Flag("raito__params", value=kwargs))



from sqlalchemy import and_, delete
from sqlalchemy.dialects.sqlite import insert

from .sqlalchemy import SQLAlchemyRoleProvider, roles_table

__all__ = ("SQLiteRoleProvider",)


class SQLiteRoleProvider(SQLAlchemyRoleProvider):
    """SQLite-based role provider.

    Required packages :code:`sqlalchemy[asyncio]`, :code:`aiosqlite` package installed (:code:`pip install raito[sqlite]`)
    """

    async def set_role(self, bot_id: int, user_id: int, role_slug: str) -> None:
        """Set the role for a specific user.

        :param bot_id: The Telegram bot ID
        :param user_id: The Telegram user ID
        :param role_slug: The role slug to assign
        """
        async with self.session_factory() as session:
            query = insert(roles_table).values(
                bot_id=bot_id,
                user_id=user_id,
                role=role_slug,
            )
            query = query.on_conflict_do_update(
                index_elements=["bot_id", "user_id"],
                set_={"role": role_slug},
            )
            await session.execute(query)
            await session.commit()

    async def remove_role(self, bot_id: int, user_id: int) -> None:
        """Remove the role for a specific user.

        :param bot_id: The Telegram bot ID
        :param user_id: The Telegram user ID
        """
        async with self.session_factory() as session:
            query = delete(roles_table).where(
                and_(
                    roles_table.c.bot_id == bot_id,
                    roles_table.c.user_id == user_id,
                ),
            )
            await session.execute(query)
            await session.commit()



from typing import TYPE_CHECKING, Literal, overload

__all__ = (
    "get_postgresql_storage",
    "get_sqlite_storage",
)

if TYPE_CHECKING:
    from aiogram.fsm.storage.redis import RedisStorage

    from .postgresql import PostgreSQLStorage
    from .sqlite import SQLiteStorage


@overload
def get_sqlite_storage(*, throw: Literal[True] = True) -> type["SQLiteStorage"]: ...
@overload
def get_sqlite_storage(*, throw: Literal[False]) -> type["SQLiteStorage"] | None: ...
def get_sqlite_storage(*, throw: bool = True) -> type["SQLiteStorage"] | None:
    try:
        from .sqlite import SQLiteStorage
    except ImportError as exc:
        if not throw:
            return None

        msg = "SQLiteStorage requires :code:`aiosqlite` package. Install it using :code:`pip install raito[sqlite]`"
        raise ImportError(msg) from exc

    return SQLiteStorage


@overload
def get_postgresql_storage(*, throw: Literal[True] = True) -> type["PostgreSQLStorage"]: ...
@overload
def get_postgresql_storage(*, throw: Literal[False]) -> type["PostgreSQLStorage"] | None: ...
def get_postgresql_storage(*, throw: bool = True) -> type["PostgreSQLStorage"] | None:
    try:
        from .postgresql import PostgreSQLStorage
    except ImportError as exc:
        if not throw:
            return None

        msg = "PostgreSQLStorage requires :code:`asyncpg`, :code:`sqlalchemy` package. Install it using :code:`pip install raito[postgresql]`"
        raise ImportError(msg) from exc

    return PostgreSQLStorage


@overload
def get_redis_storage(*, throw: Literal[True] = True) -> type["RedisStorage"]: ...
@overload
def get_redis_storage(*, throw: Literal[False]) -> type["RedisStorage"] | None: ...
def get_redis_storage(*, throw: bool = True) -> type["RedisStorage"] | None:
    try:
        from aiogram.fsm.storage.redis import RedisStorage
    except ImportError as exc:
        if not throw:
            return None

        msg = "RedisStorage requires :code:`redis` package. Install it using :code:`pip install raito[redis]`"
        raise ImportError(msg) from exc

    return RedisStorage



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any

from sqlalchemy import URL
from sqlalchemy.dialects.sqlite import insert
from sqlalchemy.ext.asyncio import create_async_engine

from .sqlalchemy import SQLAlchemyStorage, storage_table

if TYPE_CHECKING:
    from aiogram.filters.state import StateType
    from aiogram.fsm.storage.base import StorageKey

__all__ = ("SQLiteStorage",)


class SQLiteStorage(SQLAlchemyStorage):
    """SQLite storage for FSM.

    Required packages :code:`sqlalchemy[asyncio]`, :code:`aiosqlite` package installed (:code:`pip install raito[sqlite]`)
    """

    def __init__(self, url: str | URL) -> None:
        """Initialize SQLite storage.

        :param url: SQLite database URL
        :type url: str | URL
        """
        self.url = url
        engine = create_async_engine(url=self.url)
        super().__init__(engine=engine)

    async def set_state(self, key: StorageKey, state: StateType | None = None) -> None:
        """Set state for specified key.

        :param key: Storage key
        :type key: StorageKey
        :param state: New state
        :type state: StateType | None
        """
        str_key = self._build_key(key)
        async with self.session_factory() as session:
            query = insert(storage_table).values(
                key=str_key,
                state=state,
                data={},
            )
            query = query.on_conflict_do_update(
                index_elements=["key"],
                set_={"state": state, "updated_at": datetime.now(timezone.utc)},
            )
            await session.execute(query)
            await session.commit()

    async def set_data(self, key: StorageKey, data: dict[str, Any]) -> None:
        """Write data (replace).

        :param key: Storage key
        :type key: StorageKey
        :param data: New data
        :type data: Dict[str, Any]
        """
        str_key = self._build_key(key)
        async with self.session_factory() as session:
            query = insert(storage_table).values(
                key=str_key,
                data=data,
            )
            query = query.on_conflict_do_update(
                index_elements=["key"],
                set_={"data": data, "updated_at": datetime.now(timezone.utc)},
            )
            await session.execute(query)
            await session.commit()



from dataclasses import dataclass

__all__ = ("RoleData",)


@dataclass
class RoleData:
    slug: str
    name: str
    description: str
    emoji: str

    @property
    def label(self) -> str:
        return f"{self.emoji} {self.name}"



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any

from aiogram.fsm.storage.base import BaseStorage
from sqlalchemy import (
    JSON,
    Column,
    DateTime,
    Integer,
    MetaData,
    String,
    Table,
    select,
)
from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    AsyncSession,
    async_sessionmaker,
    create_async_engine,
)

if TYPE_CHECKING:
    from aiogram.fsm.storage.base import StorageKey

__all__ = ("SQLAlchemyStorage",)

metadata = MetaData()

storage_table = Table(
    "raito__fsm_storage",
    metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("key", String(255), nullable=False, unique=True, index=True),
    Column("state", String(255), nullable=True),
    Column("data", JSON, nullable=False, default={}),
    Column("created_at", DateTime, default=datetime.now(timezone.utc), nullable=False),
    Column(
        "updated_at",
        DateTime,
        default=datetime.now(timezone.utc),
        onupdate=datetime.now(timezone.utc),
        nullable=False,
    ),
)


class SQLAlchemyStorage(BaseStorage):
    """SQLAlchemy storage for FSM."""

    def __init__(
        self,
        engine: AsyncEngine,
        session_factory: async_sessionmaker[AsyncSession] | None = None,
        key_separator: str = ":",
    ) -> None:
        """Initialize SQLAlchemyStorage.

        :param engine: SQLAlchemy async engine
        :type engine: AsyncEngine
        :param session_factory: Optional session factory
        :type session_factory: async_sessionmaker[AsyncSession] | None
        :param key_separator: Separator for key parts
        :type key_separator: str
        """
        self.engine = engine
        self.session_factory = session_factory or async_sessionmaker(engine, expire_on_commit=False)
        self.key_separator = key_separator

    @classmethod
    async def from_url(
        cls,
        url: str,
        *,
        echo: bool = False,
        pool_size: int = 10,
        max_overflow: int = 0,
        **kwargs: Any,  # noqa: ANN401
    ) -> SQLAlchemyStorage:
        """Create storage from database URL.

        :param url: Database URL
        :type url: str
        :param echo: Enable SQL logging
        :type echo: bool
        :param pool_size: Connection pool size
        :type pool_size: int
        :param max_overflow: Max overflow connections
        :type max_overflow: int
        :return: Configured storage instance
        :rtype: BaseSQLAlchemyStorage
        """
        engine = create_async_engine(
            url,
            echo=echo,
            pool_size=pool_size,
            max_overflow=max_overflow,
        )
        storage = cls(engine, **kwargs)
        await storage.migrate()
        return storage

    def _build_key(self, key: StorageKey) -> str:
        """Build string key from StorageKey.

        :param key: Storage key
        :type key: StorageKey
        :return: String representation
        :rtype: str
        """
        parts = [str(key.bot_id), str(key.chat_id), str(key.user_id)]
        if key.thread_id:
            parts.append(str(key.thread_id))
        if key.business_connection_id:
            parts.append(str(key.business_connection_id))
        if key.destiny:
            parts.append(key.destiny)
        return self.key_separator.join(parts)

    async def get_state(self, key: StorageKey) -> Any | None:  # noqa: ANN401
        """Get key state.

        :param key: Storage key
        :type key: StorageKey
        :return: Current state
        :rtype: str | None
        """
        str_key = self._build_key(key)
        async with self.session_factory() as session:
            query = select(storage_table.c.state).where(storage_table.c.key == str_key)
            result = await session.execute(query)
            return result.scalar_one_or_none()

    async def get_data(self, key: StorageKey) -> dict[str, Any]:
        """Get current data for key.

        :param key: Storage key
        :type key: StorageKey
        :return: Current data
        :rtype: Dict[str, Any]
        """
        str_key = self._build_key(key)
        async with self.session_factory() as session:
            query = select(storage_table.c.data).where(storage_table.c.key == str_key)
            result = await session.execute(query)
            return result.scalar_one_or_none() or {}

    async def update_data(self, key: StorageKey, data: dict[str, Any]) -> dict[str, Any]:
        """Update data in the storage for key.

        :param key: Storage key
        :type key: StorageKey
        :param data: Data to update
        :type data: Dict[str, Any]
        :return: Updated data
        :rtype: Dict[str, Any]
        """
        current_data = await self.get_data(key)
        current_data.update(data)
        await self.set_data(key, current_data)
        return current_data

    async def close(self) -> None:
        """Close the storage."""
        await self.engine.dispose()

    async def migrate(self) -> None:
        """Create tables if not exist."""
        async with self.engine.begin() as conn:
            await conn.run_sync(metadata.create_all)



from __future__ import annotations

import contextlib
from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, Any, TypeVar

from aiogram import Bot
from aiogram.dispatcher.event.handler import HandlerObject
from aiogram.dispatcher.middlewares.base import BaseMiddleware
from aiogram.exceptions import TelegramBadRequest
from aiogram.types import CallbackQuery, Message
from aiogram.types.update import Update

from raito.plugins.pagination.data import PaginationCallbackData
from raito.plugins.pagination.enums import PaginationMode
from raito.plugins.pagination.util import get_paginator

from .paginators.base import BasePaginator

if TYPE_CHECKING:
    from aiogram.types import TelegramObject

    from raito import Raito

R = TypeVar("R")


__all__ = ("PaginatorMiddleware",)


class PaginatorMiddleware(BaseMiddleware):
    """Middleware for pagination handling."""

    def __init__(self, flag_name: str = "raito__is_pagination") -> None:
        """Initialize pagination middleware.

        :param flag_name: flag name to filter
        :type flag_name: str
        """
        self.flag_name = flag_name

    def _extract_required_data(self, data: dict[str, Any]) -> tuple[Update, Raito, Bot]:
        """Extract required data from middleware context.

        :param data: middleware data
        :type data: dict[str, Any]
        :return: tuple of update, raito and bot
        :rtype: tuple[Update, Raito, Bot]
        :raises ValueError: if required data is missing
        """
        event_update = data.get("event_update")
        if not event_update:
            raise ValueError("Event update not found")

        raito = data.get("raito")
        if not raito:
            raise ValueError("Raito instance not found")

        bot = data.get("bot")
        if not bot:
            raise ValueError("Bot instance not found")

        return event_update, raito, bot

    def _validate_callback_query(self, query: CallbackQuery, event_update: Update) -> str:
        """Validate callback query data.

        :param query: callback query
        :type query: CallbackQuery
        :param event_update: update event
        :type event_update: Update
        :return: callback data string
        :rtype: str
        :raises ValueError: if callback query is invalid
        """
        if not event_update.callback_query or not event_update.callback_query.data:
            raise ValueError("Callback query data not found")

        if not isinstance(query.message, Message):
            raise ValueError("Message not accessible")

        return event_update.callback_query.data

    def _create_paginator(self, query: CallbackQuery, data: dict[str, Any]) -> BasePaginator:
        """Create paginator instance from callback data.

        :param query: callback query
        :type query: CallbackQuery
        :param data: middleware data
        :type data: dict[str, Any]
        :return: paginator instance
        :rtype: BasePaginator
        """
        event_update, raito, bot = self._extract_required_data(data)

        if not event_update.callback_query or not event_update.callback_query.data:
            raise ValueError("Callback query data not found")

        if not isinstance(query.message, Message):
            raise ValueError("Message not accessible")

        callback_data = PaginationCallbackData.unpack(event_update.callback_query.data)
        Paginator = get_paginator(PaginationMode(callback_data.mode))
        return Paginator(
            raito=raito,
            name=callback_data.name,
            chat_id=query.message.chat.id,
            bot=bot,
            from_user=query.from_user,
            existing_message=query.message,
            current_page=callback_data.current_page,
            total_pages=callback_data.total_pages,
            limit=callback_data.limit,
        )

    async def __call__(
        self,
        handler: Callable[[TelegramObject, dict[str, Any]], Awaitable[R]],
        event: TelegramObject,
        data: dict[str, Any],
    ) -> R | None:
        """Process callback_query with pagination support.

        :param handler: Next handler in the middleware chain
        :type handler: Callable
        :param event: Telegram event (Message or CallbackQuery)
        :type event: TelegramObject
        :param data: Additional data passed through the middleware chain
        :type data: dict[str, Any]
        :return: Handler result if not throttled, None if throttled
        """
        handler_object: HandlerObject | None = data.get("handler")
        if handler_object is None:
            raise RuntimeError("Handler object not found")

        is_pagination = handler_object.flags.get(self.flag_name, False)
        if not is_pagination:
            return await handler(event, data)

        if not isinstance(event, CallbackQuery):
            return await handler(event, data)

        with contextlib.suppress(TelegramBadRequest):
            await event.answer()

        paginator: BasePaginator | None = data.get("paginator")
        if not paginator:
            paginator = self._create_paginator(event, data)

        if isinstance(event.message, Message):
            paginator.chat_id = event.message.chat.id

        bot: Bot | None = data.get("bot")
        if bot is not None:
            paginator.bot = bot

        data["paginator"] = paginator
        data["page"] = paginator.current_page
        data["offset"] = (paginator.current_page - 1) * paginator.limit
        data["limit"] = paginator.limit

        return await handler(event, data)



from __future__ import annotations

import logging
import sys
from asyncio import create_task
from typing import TYPE_CHECKING

from aiogram.dispatcher.event.event import EventObserver
from aiogram.fsm.storage.memory import MemoryStorage

from raito.plugins.commands.middleware import CommandMiddleware
from raito.plugins.commands.registration import register_bot_commands
from raito.plugins.pagination import PaginationMode, PaginatorMiddleware, get_paginator
from raito.plugins.roles import (
    BaseRoleProvider,
    IRoleProvider,
    MemoryRoleProvider,
    RoleManager,
)
from raito.plugins.roles.providers import (
    get_postgresql_provider,
    get_redis_provider,
    get_sqlite_provider,
)
from raito.plugins.roles.providers.json import JSONRoleProvider
from raito.plugins.throttling.middleware import THROTTLING_MODE, ThrottlingMiddleware
from raito.utils import loggers
from raito.utils.configuration import RaitoConfiguration
from raito.utils.const import ROOT_DIR
from raito.utils.storages import (
    get_postgresql_storage,
    get_redis_storage,
    get_sqlite_storage,
)
from raito.utils.storages.json import JSONStorage

from .routers.manager import RouterManager

if TYPE_CHECKING:
    from aiogram import Bot, Dispatcher
    from aiogram.fsm.storage.base import BaseStorage
    from aiogram.types import Message, User

    from raito.plugins.roles import IRoleProvider
    from raito.utils.types import StrOrPath

__all__ = ("Raito",)


class Raito:
    """Main class for managing the Raito utilities.

    Provides router management, middleware setup, etc.
    """

    def __init__(
        self,
        dispatcher: Dispatcher,
        routers_dir: StrOrPath,
        *,
        developers: list[int] | None = None,
        production: bool = True,
        configuration: RaitoConfiguration | None = None,
        storage: BaseStorage | None = None,
    ) -> None:
        """Initialize the Raito.

        :param dispatcher: Aiogram dispatcher instance
        :type dispatcher: Dispatcher
        :param routers_dir: Directory containing router files
        :type routers_dir: StrOrPath
        :param developers: List of developer user IDs with special privileges, defaults to None
        :type developers: list[int] | None, optional
        :param production: Whether running in production mode, defaults to True
        :type production: bool, optional
        :param configuration: Configuration instance, defaults to Configuration()
        :type configuration: Configuration | None, optional
        :param storage: Aiogram storage instance for storing data, default None
        :type storage: BaseStorage | None, optional
        """
        self.dispatcher = dispatcher
        self.routers_dir = routers_dir
        self.developers = developers or []
        self.production = production
        self.configuration = configuration or RaitoConfiguration()
        self.storage = storage or MemoryStorage()

        self.router_manager = RouterManager(dispatcher)
        self.dispatcher["raito"] = self

        self._role_provider = self._get_role_provider(self.storage)
        self.role_manager = self.configuration.role_manager or RoleManager(
            self._role_provider, developers=self.developers
        )

        self.command_parameters_error = EventObserver()

    async def setup(self) -> None:
        """Set up the Raito by loading routers and starting watchdog.

        Loads all routers from the specified directory and starts file watching
        in development mode for automatic reloading.
        """
        loggers.core.info(
            "[ 🔦 Raito ] Running in %s mode",
            "production" if self.production else "development",
        )

        provider = self.role_manager.provider
        if self.production and isinstance(provider, (MemoryRoleProvider | JSONRoleProvider)):
            loggers.roles.warn(
                "Using %s. It's not recommended for production use.",
                provider.__class__.__name__,
            )
        await self.role_manager.migrate()

        self.dispatcher.callback_query.middleware(PaginatorMiddleware("raito__is_pagination"))
        self.dispatcher.message.middleware(CommandMiddleware())

        await self.router_manager.load_routers(self.routers_dir)
        await self.router_manager.load_routers(ROOT_DIR / "handlers")

        if not self.production:
            create_task(self.router_manager.start_watchdog(self.routers_dir))  # noqa: RUF006

    def add_throttling(
        self,
        rate_limit: float,
        mode: THROTTLING_MODE = "chat",
        max_size: int = 10_000,
    ) -> None:
        """Add global throttling middleware to prevent spam.

        Applies rate limiting to both messages and callback queries.

        :param rate_limit: Time in seconds between allowed requests
        :type rate_limit: float
        :param mode: Throttling mode - 'chat', 'user', or 'bot', defaults to 'chat'
        :type mode: ThrottlingMiddleware.MODE, optional
        :param max_size: Maximum cache size for throttling records, defaults to 10_000
        :type max_size: int, optional
        """
        middleware = ThrottlingMiddleware(rate_limit=rate_limit, mode=mode, max_size=max_size)
        self.dispatcher.callback_query.middleware(middleware)
        self.dispatcher.message.middleware(middleware)

    def _get_role_provider(self, storage: BaseStorage) -> IRoleProvider:
        """Get the current role provider based on storage.

        :return: Role provider instance
        :rtype: IRoleProvider
        """
        if isinstance(storage, MemoryStorage):
            return MemoryRoleProvider(storage)

        if isinstance(storage, JSONStorage):
            return JSONRoleProvider(storage)

        redis_storage = get_redis_storage(throw=False)
        if redis_storage is not None and isinstance(storage, redis_storage):
            return get_redis_provider()(storage)

        postgresql_storage = get_postgresql_storage(throw=False)
        if postgresql_storage is not None and isinstance(storage, postgresql_storage):
            return get_postgresql_provider()(storage)

        sqlite_storage = get_sqlite_storage(throw=False)
        if sqlite_storage is not None and isinstance(storage, sqlite_storage):
            return get_sqlite_provider()(storage)

        return BaseRoleProvider(storage)

    async def paginate(
        self,
        name: str,
        chat_id: int,
        bot: Bot,
        from_user: User,
        *,
        existing_message: Message | None = None,
        mode: PaginationMode = PaginationMode.INLINE,
        current_page: int = 1,
        total_pages: int | None = None,
        limit: int = 20,
    ) -> None:
        Paginator = get_paginator(mode)
        paginator = Paginator(
            raito=self,
            name=name,
            chat_id=chat_id,
            bot=bot,
            from_user=from_user,
            existing_message=existing_message,
            current_page=current_page,
            total_pages=total_pages,
            limit=limit,
        )
        await paginator.paginate()

    async def register_commands(self, bot: Bot, locales: list[str]) -> None:
        handlers = []
        for loader in self.router_manager.loaders.values():
            handlers.extend(loader.router.message.handlers)

        await register_bot_commands(
            role_manager=self.role_manager,
            bot=bot,
            handlers=handlers,
            locales=locales,
        )

    def init_logging(self, *mute_loggers: str) -> None:
        """Configure global logging with a colored formatter.

        :param mute_loggers: List of logger names to suppress from output
        """
        logging.captureWarnings(True)

        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(loggers.ColoredFormatter())
        if mute_loggers:
            handler.addFilter(loggers.MuteLoggersFilter(*mute_loggers))

        root_logger = logging.getLogger()
        root_logger.handlers.clear()
        root_logger.addHandler(handler)
        root_logger.setLevel(logging.DEBUG if not self.production else logging.INFO)



import logging
from datetime import datetime
from typing import Literal, cast

__all__ = (
    "ColoredFormatter",
    "MuteLoggersFilter",
    "core",
    "log",
    "middlewares",
    "plugins",
    "roles",
)

LEVEL = Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]

WHITE = "\033[37m"
BRIGHT_BLACK = "\033[90m"
RESET = "\033[0m"

LEVEL_BACKGROUND_COLORS: dict[LEVEL, str] = {
    "DEBUG": "\033[42m",
    "INFO": "\033[104m",
    "WARNING": "\033[103m",
    "ERROR": "\033[101m",
    "CRITICAL": "\033[41m",
}

LEVEL_FOREGROUND_COLORS: dict[LEVEL, str] = {
    "DEBUG": "\033[32m",
    "INFO": RESET,
    "WARNING": RESET,
    "ERROR": RESET,
    "CRITICAL": "\033[31m",
}


class ColoredFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        levelname = cast(LEVEL, record.levelname)

        background_color = LEVEL_BACKGROUND_COLORS.get(levelname, "")
        foreground_color = LEVEL_FOREGROUND_COLORS.get(levelname, "")

        now = datetime.fromtimestamp(record.created).strftime("%d.%m.%Y %H:%M:%S")

        left = f"{BRIGHT_BLACK}{now}{RESET} {WHITE}{record.name}{RESET}"
        tab = " " * (64 - len(left))

        tag = f"{background_color} {levelname[0]} {RESET}"
        message = f"{foreground_color}{record.getMessage()}{RESET}"

        return f"{left}{tab} {tag} {message}"


class MuteLoggersFilter(logging.Filter):
    def __init__(self, *names: str) -> None:
        self.names = set(names)
        super().__init__()

    def filter(self, record: logging.LogRecord) -> bool:
        return record.name not in self.names


core = logging.getLogger("raito.core")
routers = logging.getLogger("raito.core.routers")
commands = logging.getLogger("raito.core.commands")

middlewares = logging.getLogger("raito.middlewares")
plugins = logging.getLogger("raito.plugins")
roles = logging.getLogger("raito.plugins.roles")

utils = logging.getLogger("raito.utils")
storages = logging.getLogger("raito.utils.storages")

log = logging.getLogger()



from __future__ import annotations

from asyncio import sleep
from typing import TYPE_CHECKING

from aiogram import Router, html

from raito.plugins.commands import description, hidden, params
from raito.plugins.roles.roles import DEVELOPER
from raito.utils.filters import RaitoCommand

if TYPE_CHECKING:
    from aiogram.types import Message

    from raito.core.raito import Raito

router = Router(name="raito.management.reload")


@router.message(RaitoCommand("reload"), DEVELOPER)
@description("Reloads a router by name")
@params(name=str)
@hidden
async def reload_router(message: Message, raito: Raito, name: str) -> None:
    router_loader = raito.router_manager.loaders.get(name)
    if not router_loader:
        await message.answer(
            f"🔎 Router <b>{html.bold(name)}</b> not found",
            parse_mode="HTML",
        )
        return

    msg = await message.answer(
        f"📦 Reloading router <b>{html.bold(name)}</b>...",
        parse_mode="HTML",
    )
    await router_loader.reload()
    await sleep(0.5)
    await msg.edit_text(f"✅ Router {html.bold(name)} reloaded", parse_mode="HTML")



from __future__ import annotations

from typing import TYPE_CHECKING, NamedTuple

from aiogram import Router, html

from raito.plugins.commands import description, hidden
from raito.plugins.roles import DEVELOPER
from raito.utils.ascii import AsciiTree, TreeNode
from raito.utils.configuration import RouterListStyle
from raito.utils.const import ROOT_DIR
from raito.utils.filters import RaitoCommand

if TYPE_CHECKING:
    from aiogram.types import Message

    from raito.core.raito import Raito
    from raito.core.routers.loader import RouterLoader

router = Router(name="raito.management.list")


class Emojis(NamedTuple):
    """Emojis for router status."""

    enabled: str
    restarting: str
    disabled: str
    not_found: str


@router.message(RaitoCommand("routers"), DEVELOPER)
@description("Lists all routers")
@hidden
async def list_routers(message: Message, raito: Raito) -> None:
    match raito.configuration.router_list_style:
        case RouterListStyle.CIRCLES:
            emojis = Emojis("🟢", "🟡", "🔴", "⚪")
        case RouterListStyle.DIAMONDS:
            emojis = Emojis("🔹", "🔸", "🔸", "🔸")
        case RouterListStyle.DIAMONDS_REVERSED:
            emojis = Emojis("🔸", "🔹", "🔹", "🔹")
        case _:
            emojis = Emojis("🟩", "🟨", "🟥", "⬜")

    def extract_loader_path(loader: RouterLoader) -> str:
        return (
            loader.path.as_posix()
            .replace(ROOT_DIR.parent.as_posix(), "")
            .replace(".py", "")
            .strip("/")
        )

    paths = {
        extract_loader_path(loader): loader for loader in raito.router_manager.loaders.values()
    }

    def get_status_icon(path: str) -> str:
        loader = paths.get(path)
        if loader and loader.is_restarting:
            return emojis.restarting
        if loader and loader.is_loaded:
            return emojis.enabled
        return emojis.disabled

    root = TreeNode("routers", is_folder=True)
    for path in paths:
        parts = path.split("/")
        current = root
        for i, part in enumerate(parts):
            full_path = "/".join(parts[: i + 1])
            is_folder = i != len(parts) - 1
            icon = get_status_icon(full_path) if not is_folder else ""
            current = current.add_child(part, prefix=icon, is_folder=is_folder)

    tree = AsciiTree().render(root)
    text = (
        html.bold("Here is your routers:")
        + "\n\n"
        + tree
        + "\n\n"
        + html.pre_language((f"{emojis[0]} — Enabled\n{emojis[2]} — Disabled\n"), "Specification")
    )

    await message.answer(text, parse_mode="HTML")



from .constraint import RoleConstraint
from .filter import RoleFilter

__all__ = (
    "ADMINISTRATOR",
    "AVAILABLE_ROLES",
    "AVAILABLE_ROLES_BY_SLUG",
    "DEVELOPER",
    "GUEST",
    "MANAGER",
    "MODERATOR",
    "OWNER",
    "SPONSOR",
    "SUPPORT",
    "TESTER",
)


def _create_role(slug: str, name: str, description: str, emoji: str) -> RoleConstraint:
    return RoleConstraint(
        RoleFilter(
            slug=slug,
            name=name,
            description=description,
            emoji=emoji,
        )
    )


DEVELOPER = _create_role(
    slug="developer",
    name="Developer",
    description="Has full access to all internal features, including debug tools and unsafe operations.",
    emoji="🖥️",
)

OWNER = _create_role(
    slug="owner",
    name="Owner",
    description="Top-level administrator with permissions to manage administrators and global settings.",
    emoji="👑",
)

ADMINISTRATOR = _create_role(
    slug="administrator",
    name="Administrator",
    description="Can manage users, moderate content, and configure most system settings.",
    emoji="💼",
)

MODERATOR = _create_role(
    slug="moderator",
    name="Moderator",
    description="Can moderate user activity, issue warnings, and enforce rules within their scope.",
    emoji="🛡️",
)

MANAGER = _create_role(
    slug="manager",
    name="Manager",
    description="Oversees non-technical operations like campaigns, tasks, or content planning.",
    emoji="📊",
)

SPONSOR = _create_role(
    slug="sponsor",
    name="Sponsor",
    description="Supporter of the project. Usually does not have administrative privileges.",
    emoji="❤️",
)

GUEST = _create_role(
    slug="guest",
    name="Guest",
    description="Has temporary access to specific internal features (e.g., analytics). Typically used for invited external users.",
    emoji="👤",
)

SUPPORT = _create_role(
    slug="support",
    name="Support",
    description="Handles user support requests and assists with onboarding or issues.",
    emoji="💬",
)

TESTER = _create_role(
    slug="tester",
    name="Tester",
    description="Helps test new features and provide feedback. May have access to experimental tools.",
    emoji="🧪",
)

AVAILABLE_ROLES = [
    i.filter.data
    for i in [
        ADMINISTRATOR,
        DEVELOPER,
        GUEST,
        MANAGER,
        MODERATOR,
        OWNER,
        SPONSOR,
        SUPPORT,
        TESTER,
    ]
]
AVAILABLE_ROLES_BY_SLUG = {role.slug: role for role in AVAILABLE_ROLES}



from __future__ import annotations

from typing import TYPE_CHECKING, overload

from .base import BaseRoleProvider
from .json import JSONRoleProvider
from .memory import MemoryRoleProvider
from .protocol import IRoleProvider
from .sql import get_postgresql_provider, get_sqlite_provider

if TYPE_CHECKING:
    from .redis import RedisRoleProvider

__all__ = (
    "BaseRoleProvider",
    "IRoleProvider",
    "JSONRoleProvider",
    "MemoryRoleProvider",
    "get_postgresql_provider",
    "get_redis_provider",
    "get_sqlite_provider",
)


@overload
def get_redis_provider() -> type[RedisRoleProvider]: ...
@overload
def get_redis_provider(*, throw: bool = False) -> type[RedisRoleProvider] | None: ...
def get_redis_provider(*, throw: bool = True) -> type[RedisRoleProvider] | None:
    try:
        from .redis import RedisRoleProvider
    except ImportError as exc:
        if not throw:
            return None

        msg = (
            "RedisRoleProvider requires redis package. Install it using `pip install raito[redis]`"
        )
        raise ImportError(msg) from exc

    return RedisRoleProvider



from __future__ import annotations

from asyncio import sleep
from pathlib import Path
from typing import TYPE_CHECKING

from .base_router import BaseRouter
from .parser import RouterParser

if TYPE_CHECKING:
    from aiogram import Dispatcher, Router

    from raito.utils.types import StrOrPath


__all__ = ("RouterLoader",)


class RouterLoader(BaseRouter, RouterParser):
    """A class for loading, unloading and reloading routers dynamically."""

    def __init__(
        self,
        name: str,
        path: StrOrPath,
        dispatcher: Dispatcher,
        router: Router | None = None,
    ) -> None:
        """Initialize RouterLoader.

        :param name: Unique name of the router
        :type name: str
        :param path: Path to the router file
        :type path: StrOrPath
        :param dispatcher: Aiogram dispatcher
        :type dispatcher: Dispatcher
        :param router: Router instance, defaults to None
        :type router: Router | None, optional
        """
        super().__init__(router)

        self.name = name
        self.path = Path(path)

        self._dispatcher = dispatcher

        self._router: Router | None = router
        self._parent_router: Router | None = None
        self._sub_routers: list[Router] = []

        self._is_restarting: bool = False
        self._is_loaded: bool = False

    @property
    def is_loaded(self) -> bool:
        """Check whether the router is currently loaded.

        :return: True if the router has been loaded and registered into the dispatcher
        :rtype: bool
        """
        return self._is_loaded

    @property
    def is_restarting(self) -> bool:
        """Check whether the router is currently being reloaded.

        :return: True if a reload operation is in progress
        :rtype: bool
        """
        return self._is_restarting

    @property
    def router(self) -> Router:
        """Get or load the router instance.

        :return: The router instance
        :rtype: Router
        """
        if self._router is None:
            self._router = self.extract_router(self.path)
            if not hasattr(self._router, "name"):
                self._router.name = self.name
        return self._router

    def load(self) -> None:
        """Load and register the router."""
        if router := self.router:
            if self._parent_router:
                self._link_to_parent(self._parent_router)
            self._dispatcher.include_router(router)
        self._is_loaded = True

    def unload(self) -> None:
        """Unload and unregister the router."""
        if self.router:
            self._unlink_from_parent()
            self._router = None
        self._is_loaded = False

    async def reload(self, timeout: float | None = None) -> None:
        """Reload the router with optional delay.

        :param timeout: Delay in seconds before reloading, defaults to None
        :type timeout: float | None, optional
        """
        if not self._is_restarting:
            self._is_restarting = True
            self.unload()

            if timeout:
                await sleep(timeout)

            self.load()
            self._is_restarting = False



from __future__ import annotations

import re
from collections.abc import Sequence
from typing import Any

from aiogram.filters import Command, CommandObject
from aiogram.filters.command import CommandException, CommandPatternType
from aiogram.utils.magic_filter import MagicFilter

PREFIX = ".rt "

__all__ = ("RaitoCommand",)


class RaitoCommand(Command):
    """A filter for Raito bot commands.

    This class filters messages that match the Raito command format:
    ".rt <command> [arguments]"

    The filter matches commands exactly and optionally allows additional arguments.
    Commands are case-sensitive and must match the prefix ".rt" followed by
    one of the specified command strings.

    Example:

    .. code-block:: python

        @router.message(RaitoCommand("test"))
        async def test(message: Message):
            # Handles messages like:
            # ".rt test"
            # ".rt test foo bar 123"
            pass

    """

    def __init__(
        self,
        *values: CommandPatternType,
        commands: Sequence[CommandPatternType] | CommandPatternType | None = None,
        ignore_case: bool = False,
        magic: MagicFilter | None = None,
    ) -> None:
        """Initialize the RaitoCommand filter.

        :param commands: One or more command strings to match
        :param ignore_case: Ignore case (Does not work with regexp, use flags instead)
        :param magic: Validate command object via Magic filter after all checks done
        :raises ValueError: If no commands are specified
        """
        super().__init__(
            *values,
            commands=commands,
            ignore_case=ignore_case,
            ignore_mention=True,
            magic=magic,
        )

        self.prefix = PREFIX
        pattern = (
            rf"^{re.escape(self.prefix)} (?:{'|'.join(map(re.escape, self.commands))})(?: .+)?$"
        )
        self._regex = re.compile(pattern)

    def extract_command(self, text: str) -> CommandObject:
        # First step: separate command with arguments
        # ".rt command arg1 arg2" -> ".rt", "command", ["arg1 arg2"]
        try:
            prefix, command, *args = text.split(maxsplit=2)
        except ValueError as exc:
            msg = "Not enough values to unpack"
            raise CommandException(msg) from exc

        return CommandObject(prefix=prefix + " ", command=command, args=args[0] if args else None)

    def update_handler_flags(self, flags: dict[str, Any]) -> None:
        super().update_handler_flags(flags)
        flags["raito__hidden"] = True



from __future__ import annotations

from typing import TYPE_CHECKING

from aiogram import F, Router, html
from aiogram.fsm.state import State, StatesGroup

from raito.plugins.commands import description, hidden
from raito.plugins.roles.roles import ADMINISTRATOR, DEVELOPER, OWNER
from raito.utils.filters import RaitoCommand

if TYPE_CHECKING:
    from aiogram.fsm.context import FSMContext
    from aiogram.types import Message

    from raito.core.raito import Raito

router = Router(name="raito.roles.revoke")


class RevokeRoleGroup(StatesGroup):
    """State group for revoking roles."""

    user_id = State()


@router.message(RaitoCommand("revoke"), DEVELOPER | OWNER | ADMINISTRATOR)
@description("Revokes a role from a user")
@hidden
async def revoke(message: Message, state: FSMContext) -> None:
    await message.answer("👤 Enter user ID:")
    await state.set_state(RevokeRoleGroup.user_id)


@router.message(
    RevokeRoleGroup.user_id,
    F.text and F.text.isdigit(),
    DEVELOPER | OWNER | ADMINISTRATOR,
)
async def revoke_role(message: Message, raito: Raito, state: FSMContext) -> None:
    if not message.bot:
        await message.answer("🚫 Bot not found")
        return
    if not message.text or not message.text.isdigit():
        await message.answer("🚫 Invalid user ID")
        return
    if not message.from_user:
        await message.answer("🚫 Initiator not found")
        return
    await state.set_state()

    role_slug = await raito.role_manager.get_role(
        message.bot.id,
        int(message.text),
    )
    if not role_slug:
        await message.answer("⚠️ User does not have the role")
        return

    try:
        await raito.role_manager.revoke_role(
            message.bot.id,
            message.from_user.id,
            int(message.text),
        )
    except PermissionError:
        await message.answer("🚫 Permission denied")
        return

    role = raito.role_manager.get_role_data(role_slug)
    await message.answer(f"🛑 User revoked from {html.bold(role.label)}", parse_mode="HTML")



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from aiogram import html

if TYPE_CHECKING:
    from aiogram.filters.command import CommandObject

EXAMPLE_VALUES = {
    bool: "yes",
    str: "word",
    int: "10",
    float: "3.14",
}


def get_command_help(
    command_object: CommandObject,
    params: dict[str, type[Any]],
    description: str | None = None,
) -> str:
    """Get the help message of a command."""
    cmd = command_object.prefix + command_object.command

    signature = cmd
    example = cmd
    for param, value_type in params.items():
        signature += f" [{param}]"
        example += " " + EXAMPLE_VALUES[value_type]

    description = "\n" + html.expandable_blockquote(html.italic(description)) if description else ""

    return (
        html.bold(cmd)
        + description
        + html.italic("\n\n— Signature:\n")
        + html.code(signature)
        + html.italic("\n\n— Example:\n")
        + html.code(example)
    )



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from aiogram import Router

__all__ = ("BaseRouter",)


class BaseRouter:
    """Base class providing router linking and unlinking functionality."""

    def __init__(self, router: Router | None) -> None:
        """Initialize the BaseRouter instance.

        :param router: Router instance to manage
        :type router: Router | None, optional
        """
        self._router = router

    @property
    def router(self) -> Router | None:
        """Get the managed router instance.

        :return: The managed router instance or None if not set
        :rtype: Router | None
        """
        return self._router

    def _unlink_from_parent(self) -> None:
        """Unlink router from its parent router."""
        if not self._router or not self._router.parent_router:
            return

        parent = self._router.parent_router
        parent.sub_routers = [r for r in parent.sub_routers if r.name != self._router.name]

    def _link_to_parent(self, parent: Router) -> None:
        """Link router to a parent router.

        :param parent: Parent router to link to
        :type parent: Router
        """
        if self._router and self._router not in parent.sub_routers:
            parent.sub_routers.append(self._router)



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, overload

__all__ = (
    "get_postgresql_provider",
    "get_sqlite_provider",
)

if TYPE_CHECKING:
    from .postgresql import PostgreSQLRoleProvider
    from .sqlite import SQLiteRoleProvider


@overload
def get_sqlite_provider(*, throw: Literal[True] = True) -> type[SQLiteRoleProvider]: ...
@overload
def get_sqlite_provider(*, throw: Literal[False]) -> type[SQLiteRoleProvider] | None: ...
def get_sqlite_provider(*, throw: bool = True) -> type[SQLiteRoleProvider] | None:
    try:
        from .sqlite import SQLiteRoleProvider
    except ImportError as exc:
        if not throw:
            return None

        msg = "SQLiteRoleProvider requires :code:`aiosqlite` package. Install it using :code:`pip install raito[sqlite]`"
        raise ImportError(msg) from exc

    return SQLiteRoleProvider


@overload
def get_postgresql_provider(*, throw: Literal[True] = True) -> type[PostgreSQLRoleProvider]: ...
@overload
def get_postgresql_provider(*, throw: Literal[False]) -> type[PostgreSQLRoleProvider] | None: ...
def get_postgresql_provider(*, throw: bool = True) -> type[PostgreSQLRoleProvider] | None:
    try:
        from .postgresql import PostgreSQLRoleProvider
    except ImportError as exc:
        if not throw:
            return None

        msg = "PostgreSQLRoleProvider requires :code:`asyncpg`, :code:`sqlalchemy` package. Install it using :code:`pip install raito[postgresql]`"
        raise ImportError(msg) from exc

    return PostgreSQLRoleProvider



from sqlalchemy import (
    BigInteger,
    Column,
    Index,
    Integer,
    MetaData,
    String,
    Table,
    and_,
    select,
)
from sqlalchemy.ext.asyncio import (
    AsyncSession,
    async_sessionmaker,
)

from raito.plugins.roles.providers.protocol import IRoleProvider
from raito.utils.storages.sql.sqlalchemy import SQLAlchemyStorage

__all__ = ("SQLAlchemyRoleProvider",)

metadata = MetaData()

roles_table = Table(
    "raito__user_roles",
    metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("bot_id", BigInteger, nullable=False),
    Column("user_id", BigInteger, nullable=False),
    Column("role", String, nullable=False),
    Index("idx_bot_user", "bot_id", "user_id", unique=True),
)


class SQLAlchemyRoleProvider(IRoleProvider):
    """Base SQLAlchemy role provider."""

    def __init__(
        self,
        storage: SQLAlchemyStorage,
        session_factory: async_sessionmaker[AsyncSession] | None = None,
    ) -> None:
        """Initialize SQLAlchemyRoleProvider.

        :param engine: SQLAlchemy async engine
        :param session_factory: Optional session factory, defaults to None
        """
        self.storage = storage
        self.engine = self.storage.engine
        self.session_factory = session_factory or async_sessionmaker(
            self.engine,
            class_=AsyncSession,
            expire_on_commit=False,
        )

    async def get_role(self, bot_id: int, user_id: int) -> str | None:
        """Get the role for a specific user.

        :param bot_id: The Telegram bot ID
        :param user_id: The Telegram user ID
        :return: The role slug or None if not found
        """
        async with self.session_factory() as session:
            query = select(roles_table.c.role).where(
                and_(
                    roles_table.c.bot_id == bot_id,
                    roles_table.c.user_id == user_id,
                ),
            )
            result = await session.execute(query)
            return result.scalar_one_or_none()

    async def migrate(self) -> None:
        """Initialize the storage backend (create tables, etc.)."""
        async with self.engine.begin() as conn:
            await conn.run_sync(metadata.create_all)

    async def close(self) -> None:
        """Close the database connection."""
        await self.engine.dispose()

    async def get_users(self, bot_id: int, role_slug: str) -> list[int]:
        """Get all users with a specific role.

        :param bot_id: The Telegram bot ID
        :param role_slug: The role slug to check for
        :return: A list of Telegram user IDs
        """
        async with self.session_factory() as session:
            query = select(roles_table.c.user_id).where(
                and_(
                    roles_table.c.bot_id == bot_id,
                    roles_table.c.role == role_slug,
                )
            )
            result = await session.execute(query)
            return [row[0] for row in result.all()]



from __future__ import annotations

from typing import TYPE_CHECKING, cast
from uuid import uuid4

from aiogram.types import CallbackQuery, Update
from aiogram.utils.keyboard import InlineKeyboardBuilder

from raito.plugins.pagination.data import PaginationCallbackData
from raito.plugins.pagination.enums import PaginationMode
from raito.plugins.pagination.paginators.protocol import IPaginator
from raito.utils.configuration import RaitoConfiguration

if TYPE_CHECKING:
    from aiogram import Bot
    from aiogram.types import (
        InlineKeyboardMarkup,
        Message,
        User,
    )

    from raito.core.raito import Raito

__all__ = ("BasePaginator",)


class BasePaginator(IPaginator):
    """Base paginator class."""

    def __init__(
        self,
        raito: Raito,
        name: str,
        chat_id: int,
        bot: Bot,
        from_user: User,
        *,
        existing_message: Message | None = None,
        current_page: int = 1,
        total_pages: int | None = None,
        limit: int = 20,
    ) -> None:
        """Initialize the Raito.

        :param raito: raito instance
        :type raito: Raito
        :param name: pagination name
        :type name: str
        :param chat_id: target chat id
        :type chat_id: int
        :param bot: bot instance
        :type bot: Bot
        :param from_user: user who triggered pagination
        :type from_user: User
        :param existing_message: existing message to edit
        :type existing_message: Message | None
        :param current_page: current page number
        :type current_page: int
        :param total_pages: total pages count, defaults None
        :type total_pages: int | None
        :param limit: items per page
        :type limit: int
        """
        self._validate_parameters(name, current_page, total_pages, limit)

        self.raito = raito
        self.name = name
        self.chat_id = chat_id
        self.bot = bot
        self.from_user = from_user
        self.existing_message = existing_message
        self._current_page = current_page
        self._total_pages = total_pages
        self._limit = limit

        self.configuration: RaitoConfiguration = self.raito.configuration

    def _validate_parameters(
        self, name: str, current_page: int, total_pages: int | None, limit: int
    ) -> None:
        """Validate paginator parameters.

        :param name: pagination name
        :type name: str
        :param current_page: current page number
        :type current_page: int
        :param total_pages: total pages count
        :type total_pages: int | None
        :param limit: items per page
        :type limit: int
        :raises ValueError: if parameters are invalid
        """
        if not name or len(name) > 32:
            raise ValueError("Name must be 1-32 characters long")

        if current_page < 1 or current_page >= 10**10:
            raise ValueError("Current page must be between 1 and 10^10")

        if total_pages is not None and (total_pages < current_page or total_pages >= 10**10):
            raise ValueError("Total pages must be >= current page and < 10^10")

        if limit < 1 or limit >= 10**5:
            raise ValueError("Limit must be between 1 and 10^5")

    def _create_callback_query(self) -> CallbackQuery:
        """Create callback query for pagination event.

        :return: callback query instance
        :rtype: CallbackQuery
        """
        return CallbackQuery(
            id=str(uuid4()),
            from_user=self.from_user,
            chat_instance=str(uuid4()),
            message=self.existing_message,
            data=PaginationCallbackData(
                mode=self.mode.value,
                name=self.name,
                current_page=self.current_page,
                total_pages=self.total_pages,
                limit=self.limit,
            ).pack(),
        )

    async def _invoke_callback(self) -> None:
        """Invoke pagination callback through dispatcher."""
        if not self.bot:
            raise ValueError("Bot instance not set")

        await self.raito.dispatcher.feed_update(
            bot=self.bot,
            update=Update(
                update_id=0,
                callback_query=self._create_callback_query(),
            ),
            paginator=self,
        )

    @property
    def current_page(self) -> int:
        """Get current page number."""
        return self._current_page

    @property
    def total_pages(self) -> int | None:
        """Get total pages count."""
        return self._total_pages

    @property
    def limit(self) -> int:
        """Get items per page limit."""
        return self._limit

    @property
    def mode(self) -> PaginationMode:
        """Get pagination mode.

        :raises NotImplementedError: must be implemented by subclasses
        """
        raise NotImplementedError("Subclasses must implement mode property")

    async def paginate(self) -> None:
        """Start pagination."""
        await self._invoke_callback()

    def get_previous_page(self) -> int:
        """Get previous page number with loop support.

        :return: previous page number
        :rtype: int
        """
        previous_page = self.current_page - 1

        if previous_page >= 1:
            return previous_page

        if self.configuration.pagination_style.loop_navigation and self.total_pages:
            return self.total_pages

        return self.current_page

    def get_next_page(self) -> int:
        """Get next page number with loop support.

        :return: next page number
        :rtype: int
        """
        next_page = self.current_page + 1

        if not self.total_pages:
            return next_page

        if next_page <= self.total_pages:
            return next_page

        if self.configuration.pagination_style.loop_navigation:
            return 1

        return self.current_page

    def _create_page_callback_data(self, page: int) -> PaginationCallbackData:
        """Create callback data for specific page.

        :param page: target page number
        :type page: int
        :return: callback data instance
        :rtype: PaginationCallbackData
        """
        return PaginationCallbackData(
            mode=self.mode.value,
            name=self.name,
            current_page=page,
            total_pages=self.total_pages,
            limit=self.limit,
        )

    def build_navigation(self) -> InlineKeyboardMarkup:
        """Build navigation inline keyboard.

        :return: navigation keyboard markup
        :rtype: InlineKeyboardMarkup
        """
        builder = InlineKeyboardBuilder()
        adjust = []

        builder.button(
            text=self.configuration.pagination_style.controls.previous,
            callback_data=self._create_page_callback_data(self.get_previous_page()),
        )
        builder.button(
            text=self.configuration.pagination_style.controls.next,
            callback_data=self._create_page_callback_data(self.get_next_page()),
        )
        adjust.append(2)

        if self.configuration.pagination_style.show_counter:
            counter_text = self.configuration.pagination_style.text_format.counter_template.format(
                current=self.current_page, total=self.total_pages or "∞"
            )
            builder.button(text=counter_text, callback_data="rt_p__counter")
            adjust.append(1)

        return cast(InlineKeyboardBuilder, builder.adjust(*adjust)).as_markup()

    @classmethod
    def calc_total_pages(cls, total_items: int, limit: int) -> int:
        """Calculate total pages from items count.

        :param total_items: total items count
        :type total_items: int
        :param limit: items per page
        :type limit: int
        :return: total pages count
        :rtype: int
        """
        if limit <= 0:
            raise ValueError("Limit must be greater than 0")

        return (total_items + limit - 1) // limit



from asyncio import Lock

from aiogram.fsm.storage.base import BaseStorage, DefaultKeyBuilder, StorageKey

from .protocol import IRoleProvider

__all__ = ("BaseRoleProvider",)


class BaseRoleProvider(IRoleProvider):
    """Base role provider class."""

    def __init__(self, storage: BaseStorage) -> None:
        """Initialize BaseRoleProvider."""
        self.storage = storage
        self.key_builder = DefaultKeyBuilder(with_destiny=True, with_bot_id=True)
        self._lock = Lock()

    def _build_key(self, *, bot_id: int) -> StorageKey:
        """Build a storage key for a specific bot.

        :param bot_id: The Telegram bot ID
        :return: The storage key
        """
        return StorageKey(  # applies to a single bot across all chats
            bot_id=bot_id,
            chat_id=0,
            user_id=0,
            destiny="roles",
        )

    async def get_role(self, bot_id: int, user_id: int) -> str | None:
        """Get the role for a specific user.

        :param bot_id: The Telegram bot ID
        :param user_id: The Telegram user ID
        :return: The role slug or None if not found
        """
        key = self._build_key(bot_id=bot_id)
        data: dict[str, str] = await self.storage.get_data(key)
        return data.get(str(user_id))

    async def set_role(self, bot_id: int, user_id: int, role_slug: str) -> None:
        """Set the role for a specific user.

        :param bot_id: The Telegram bot ID
        :param user_id: The Telegram user ID
        :param role_slug: The role slug to assign
        """
        key = self._build_key(bot_id=bot_id)
        async with self._lock:
            data: dict[str, str] = await self.storage.get_data(key)
            data[str(user_id)] = role_slug
            await self.storage.set_data(key, data)

    async def remove_role(self, bot_id: int, user_id: int) -> None:
        """Remove the role for a specific user.

        :param bot_id: The Telegram bot ID
        :param user_id: The Telegram user ID
        """
        key = self._build_key(bot_id=bot_id)
        async with self._lock:
            data: dict[str, str] = await self.storage.get_data(key)
            data.pop(str(user_id), None)
            await self.storage.set_data(key, data)

    async def migrate(self) -> None:
        """Initialize the storage backend (create tables, etc.)."""
        return

    async def get_users(self, bot_id: int, role_slug: str) -> list[int]:
        """Get all users with a specific role.

        :param bot_id: The Telegram bot ID
        :param role_slug: The role slug to check for
        :return: A list of Telegram user IDs
        """
        key = self._build_key(bot_id=bot_id)
        async with self._lock:
            data: dict[str, str] = await self.storage.get_data(key)
            return [int(user_id) for user_id, user_role in data.items() if user_role == role_slug]



from __future__ import annotations

import json
from pathlib import Path
from typing import Any

from aiogram.fsm.state import State
from aiogram.fsm.storage.base import BaseStorage, StateType, StorageKey

from raito.utils import loggers

__all__ = ("JSONStorage",)


class JSONStorage(BaseStorage):
    """JSON-based FSM storage for development and testing.

    Stores FSM state and data in a local JSON file as a flat key-value mapping.
    """

    def __init__(self, path: str | Path, *, key_separator: str = ":") -> None:
        """Initialize JSONStorage.

        :param path: Path to the JSON file to be used for persistent storage
        :param key_separator: Delimiter used when constructing keys
        """
        self.path = Path(path)
        self.key_separator = key_separator

        self._data: dict[str, dict[str, Any]] = {}
        self._load()

    def _load(self) -> None:
        """Load JSON file contents into memory."""
        if self.path.exists():
            try:
                content = self.path.read_text(encoding="utf-8")
                self._data = json.loads(content)
            except json.JSONDecodeError as exc:
                loggers.storages.warning("JSON decode error: %s — file will be ignored", exc)
                self._data = {}
            except UnicodeDecodeError as exc:
                loggers.storages.warning("Invalid encoding in %s: %s", self.path, exc)
                self._data = {}
            except OSError as exc:
                loggers.storages.warning("Failed to read JSON storage file: %s", exc)
                self._data = {}

    def _save(self) -> None:
        """Write current memory state to JSON file."""
        self.path.write_text(
            json.dumps(self._data, ensure_ascii=False, indent=2),
            encoding="utf-8",
        )

    def _build_key(self, key: StorageKey) -> str:
        """Construct a unique key string from StorageKey.

        :param key: FSM storage key
        :return: String key
        """
        parts = [str(key.bot_id), str(key.chat_id), str(key.user_id)]
        if key.thread_id:
            parts.append(str(key.thread_id))
        if key.business_connection_id:
            parts.append(str(key.business_connection_id))
        if key.destiny:
            parts.append(key.destiny)
        return self.key_separator.join(parts)

    async def get_state(self, key: StorageKey) -> str | None:
        """Retrieve the current state for a key.

        :param key: FSM storage key
        :return: Current state or None
        """
        return self._data.get(self._build_key(key), {}).get("state")

    async def set_state(self, key: StorageKey, state: StateType | None = None) -> None:
        """Set a new state for the given key.

        :param key: FSM storage key
        :param state: New state to store
        """
        if isinstance(state, State):
            state = state.state

        str_key = self._build_key(key)
        self._data.setdefault(str_key, {})["state"] = state
        self._save()

    async def get_data(self, key: StorageKey) -> dict[str, Any]:
        """Retrieve data dictionary for the key.

        :param key: FSM storage key
        :return: Stored data or empty dict
        """
        return self._data.get(self._build_key(key), {}).get("data", {})

    async def set_data(self, key: StorageKey, data: dict[str, Any]) -> None:
        """Set data dictionary for the key.

        :param key: FSM storage key
        :param data: Data to store
        """
        str_key = self._build_key(key)
        self._data.setdefault(str_key, {})["data"] = data
        self._save()

    async def update_data(self, key: StorageKey, data: dict[str, Any]) -> dict[str, Any]:
        """Update the current data for the key.

        :param key: FSM storage key
        :param data: New data to merge with existing
        :return: Updated data
        """
        current = await self.get_data(key)
        current.update(data)

        await self.set_data(key, current)
        return current

    async def clear(self) -> None:
        """Clear all states and data from storage."""
        self._data.clear()
        self._save()

    async def close(self) -> None:
        """Close the storage (optional flush)"""
        self._save()



from __future__ import annotations

from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from typing import TYPE_CHECKING

from aiogram import Router

if TYPE_CHECKING:
    from raito.utils.types import StrOrPath


__all__ = ("RouterParser",)


class RouterParser:
    """Parses routers from Python files."""

    @classmethod
    def extract_router(cls, file_path: StrOrPath) -> Router:
        """Extract router from a Python file.

        :param file_path: Path to the Python file
        :type file_path: StrOrPath
        :return: Extracted router instance
        :rtype: Router
        """
        file_path = Path(file_path)
        module = cls._load_module(file_path)
        return cls._validate_router(module)

    @classmethod
    def _load_module(cls, file_path: StrOrPath) -> object:
        """Load module from file path.

        :param file_path: Path to the Python file to load
        :type file_path: StrOrPath
        :return: Loaded module object
        :rtype: object
        :raises ModuleNotFoundError: If module cannot be loaded from the file path
        """
        spec = spec_from_file_location("dynamic_module", file_path)

        if spec is None or spec.loader is None:
            msg = f"Cannot load module from {file_path}"
            raise ModuleNotFoundError(msg)

        module = module_from_spec(spec)
        spec.loader.exec_module(module)
        return module

    @classmethod
    def _validate_router(cls, module: object) -> Router:
        """Validate and return router from module.

        :param module: Module object to extract router from
        :type module: object
        :return: Validated router instance
        :rtype: Router
        :raises TypeError: If the module doesn't contain a valid Router instance
        """
        router = getattr(module, "router", None)
        if not isinstance(router, Router):
            msg = f"Expected Router, got {type(router).__name__}"
            raise TypeError(msg)

        return router



from enum import IntEnum, unique

__all__ = ("PaginationMode",)


@unique
class PaginationMode(IntEnum):
    """Pagination display modes.

    :cvar INLINE: inline keyboard pagination
    :cvar TEXT: text-based pagination
    :cvar PHOTO: photo pagination
    """

    INLINE = 0
    TEXT = 1
    PHOTO = 2



from dataclasses import dataclass, field
from enum import IntEnum, unique

from pydantic import BaseModel

from raito.plugins.roles.manager import RoleManager

__all__ = (
    "PaginationControls",
    "PaginationStyle",
    "PaginationTextFormat",
    "RaitoConfiguration",
    "RouterListStyle",
)


@unique
class RouterListStyle(IntEnum):
    SQUARES = 0
    CIRCLES = 1
    DIAMONDS = 2
    DIAMONDS_REVERSED = 3


@dataclass
class PaginationControls:
    previous: str = "◀️"
    next: str = "▶️"


@dataclass
class PaginationTextFormat:
    counter_template: str = "{current} / {total}"


@dataclass
class PaginationStyle:
    loop_navigation: bool = True
    controls: PaginationControls = field(default_factory=PaginationControls)
    text_format: PaginationTextFormat = field(default_factory=PaginationTextFormat)
    show_counter: bool = True


class RaitoConfiguration(BaseModel):
    router_list_style: RouterListStyle = RouterListStyle.DIAMONDS
    role_manager: RoleManager | None = None
    pagination_style: PaginationStyle = PaginationStyle()

    class Config:
        arbitrary_types_allowed = True



from aiogram.fsm.storage.redis import RedisStorage

from .base import BaseRoleProvider
from .protocol import IRoleProvider

__all__ = ("RedisRoleProvider",)


class RedisRoleProvider(BaseRoleProvider, IRoleProvider):
    """Redis-based role provider.

    Redis storage required :code:`redis` package installed (:code:`pip install raito[redis]`)
    """

    def __init__(self, storage: RedisStorage) -> None:
        """Initialize RedisRoleProvider."""
        super().__init__(storage=storage)

        self.storage: RedisStorage
        self.storage.key_builder = self.key_builder



from typing import TYPE_CHECKING, Protocol

from raito.plugins.pagination.enums import PaginationMode

if TYPE_CHECKING:
    from aiogram.types import InlineKeyboardMarkup

__all__ = ("IPaginator",)


class IPaginator(Protocol):
    """Protocol for paginator implementations.

    Defines the interface that all paginator classes must implement.
    """

    @property
    def current_page(self) -> int:
        """Get current page number.

        :return: current page number
        :rtype: int
        """
        ...

    @property
    def total_pages(self) -> int | None:
        """Get total pages count.

        :return: total pages or None
        :rtype: int | None
        """
        ...

    @property
    def limit(self) -> int:
        """Get items per page limit.

        :return: items per page
        :rtype: int
        """
        ...

    @property
    def mode(self) -> "PaginationMode":
        """Get pagination mode.

        :return: pagination mode
        :rtype: PaginationMode
        """
        ...

    async def paginate(self) -> None:
        """Start pagination."""
        ...

    def get_previous_page(self) -> int:
        """Get previous page number.

        :return: previous page number
        :rtype: int
        """
        ...

    def get_next_page(self) -> int:
        """Get next page number.

        :return: next page number
        :rtype: int
        """
        ...

    def build_navigation(self) -> "InlineKeyboardMarkup":
        """Build navigation keyboard.

        :return: navigation keyboard markup
        :rtype: InlineKeyboardMarkup
        """
        ...

    @classmethod
    def calc_total_pages(cls, total_items: int, limit: int) -> int:
        """Calculate total pages from items count.

        :param total_items: total items count
        :type total_items: int
        :param limit: items per page
        :type limit: int
        :return: total pages count
        :rtype: int
        """
        ...



from .core.raito import Raito
from .plugins import keyboards as keyboard
from .plugins.commands.flags import description, hidden, params
from .plugins.lifespan.decorator import lifespan
from .plugins.pagination import on_pagination
from .plugins.roles import (
    ADMINISTRATOR,
    DEVELOPER,
    GUEST,
    MANAGER,
    MODERATOR,
    OWNER,
    SPONSOR,
    SUPPORT,
    TESTER,
)
from .plugins.throttling.flag import limiter
from .utils.loggers import log

debug = log.debug

__all__ = (
    "ADMINISTRATOR",
    "DEVELOPER",
    "GUEST",
    "MANAGER",
    "MODERATOR",
    "OWNER",
    "SPONSOR",
    "SUPPORT",
    "TESTER",
    "Raito",
    "debug",
    "description",
    "hidden",
    "keyboard",
    "lifespan",
    "limiter",
    "log",
    "on_pagination",
    "params",
)



from aiogram.filters.callback_data import CallbackData

__all__ = ("PaginationCallbackData",)

_PREFIX = "rt_p"


class PaginationCallbackData(CallbackData, prefix=_PREFIX):  # type: ignore[call-arg]
    """Callback data for inline navigation buttons.

    :param mode: pagination mode
    :type mode: int
    :param name: paginator name
    :type name: str
    :param current_page: current page number
    :type current_page: int
    :param total_pages: total pages count or None
    :type total_pages: int | None
    :param limit: items per page
    :type limit: int
    """

    mode: int
    name: str
    current_page: int
    total_pages: int | None
    limit: int



from .json import JSONStorage
from .sql import get_postgresql_storage, get_redis_storage, get_sqlite_storage

__all__ = (
    "JSONStorage",
    "get_postgresql_storage",
    "get_redis_storage",
    "get_sqlite_storage",
)



from __future__ import annotations

from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, Any, Literal, TypeVar

from aiogram.dispatcher.event.handler import HandlerObject
from aiogram.dispatcher.middlewares.base import BaseMiddleware
from aiogram.types import CallbackQuery, Message
from cachetools import TTLCache

if TYPE_CHECKING:
    from aiogram.types import TelegramObject

R = TypeVar("R")

__all__ = (
    "THROTTLING_MODE",
    "ThrottlingMiddleware",
)

THROTTLING_MODE = Literal["user", "chat", "bot"]


class ThrottlingMiddleware(BaseMiddleware):
    """Middleware for global and per-handler throttling."""

    def __init__(
        self,
        rate_limit: float = 0.5,
        mode: THROTTLING_MODE = "chat",
        max_size: int = 10_000,
    ) -> None:
        """Initialize the middleware.

        :param rate_limit: Global throttling timeout (seconds)
        :param mode: Throttling scope — 'user', 'chat', or 'bot'
        :param max_size: Maximum number of keys stored in cache
        """
        self.rate_limit = rate_limit
        self.mode: THROTTLING_MODE = mode
        self.max_size = max_size

        self._global_cache: TTLCache[int, bool] = self._create_cache(ttl=self.rate_limit)
        self._local_cache: dict[int, TTLCache[int, bool]] = {}

    def _create_cache(self, ttl: float) -> TTLCache[int, bool]:
        return TTLCache(maxsize=self.max_size, ttl=ttl)

    def _get_key(self, event: TelegramObject, mode: THROTTLING_MODE) -> int | None:
        if isinstance(event, Message):
            if not event.from_user or not event.chat or not event.bot:
                return None
            match mode:
                case "user":
                    return event.from_user.id
                case "chat":
                    return event.chat.id
                case "bot":
                    return event.bot.id
        elif isinstance(event, CallbackQuery):
            if not event.from_user or not event.message or not event.bot:
                return None
            match mode:
                case "user":
                    return event.from_user.id
                case "chat":
                    return event.message.chat.id
                case "bot":
                    return event.bot.id
        return None

    async def _global_throttling(
        self,
        handler: Callable[[TelegramObject, dict[str, Any]], Awaitable[R]],
        event: TelegramObject,
        data: dict[str, Any],
    ) -> R | None:
        global_key = self._get_key(event, self.mode)
        if global_key is None:
            return await handler(event, data)

        if global_key in self._global_cache:
            return None

        self._global_cache[global_key] = True
        return await handler(event, data)

    async def _local_throttling(
        self,
        handler: Callable[[TelegramObject, dict[str, Any]], Awaitable[R]],
        event: TelegramObject,
        data: dict[str, Any],
        *,
        rate: float,
        mode: THROTTLING_MODE,
        handler_id: int,
    ) -> R | None:
        entity_id = self._get_key(event, mode)
        if entity_id is None:
            return await handler(event, data)

        cache = self._local_cache.setdefault(handler_id, self._create_cache(ttl=rate))
        if entity_id in cache:
            return None

        cache[entity_id] = True
        return await handler(event, data)

    async def __call__(
        self,
        handler: Callable[[TelegramObject, dict[str, Any]], Awaitable[R]],
        event: TelegramObject,
        data: dict[str, Any],
    ) -> R | None:
        handler_object: HandlerObject | None = data.get("handler")
        if handler_object is None:
            return await handler(event, data)

        limiter_data: dict[str, Any] = handler_object.flags.get("raito__limiter", {})
        local_rate: float | None = limiter_data.get("rate_limit")
        local_mode: THROTTLING_MODE | None = limiter_data.get("mode")

        if local_rate is not None and local_mode is not None:
            handler_id = id(handler_object.callback)
            return await self._local_throttling(
                handler,
                event,
                data,
                rate=local_rate,
                mode=local_mode,
                handler_id=handler_id,
            )
        else:
            return await self._global_throttling(handler, event, data)



import inspect
from collections.abc import Callable
from functools import partial, update_wrapper
from typing import Any, ParamSpec, TypeVar, cast

P = ParamSpec("P")
R = TypeVar("R")


def safe_partial(func: Callable[P, R], /, **kwargs: Any) -> Callable[P, R]:  # noqa: ANN401
    """
    Creates a partial version of a function, keeping only keyword arguments
    that are accepted by the original function.

    :param func: The original function to partially apply
    :param kwargs: Keyword arguments to bind
    :return: A new callable with partially applied arguments
    """
    signature = inspect.signature(func)

    valid_parameters = {
        name
        for name, param in signature.parameters.items()
        if param.kind in (param.POSITIONAL_OR_KEYWORD, param.KEYWORD_ONLY)
    }

    filtered_kwargs = {k: v for k, v in kwargs.items() if k in valid_parameters}
    wrapped = partial(func, **filtered_kwargs)
    return cast(Callable[P, R], update_wrapper(wrapped, func))



from collections.abc import Callable
from functools import wraps
from typing import Concatenate, Literal, TypeAlias, cast, overload

from aiogram.types import InlineKeyboardMarkup, ReplyKeyboardMarkup
from aiogram.utils.keyboard import InlineKeyboardBuilder, ReplyKeyboardBuilder
from typing_extensions import ParamSpec, TypeVar

__all__ = ("dynamic_keyboard",)


P = ParamSpec("P")
BuilderT = TypeVar("BuilderT", InlineKeyboardBuilder, ReplyKeyboardBuilder)

ButtonData: TypeAlias = str | tuple[str] | tuple[str, str] | list[str]
KeyboardMarkupT: TypeAlias = InlineKeyboardMarkup | ReplyKeyboardMarkup

InlineFn: TypeAlias = Callable[Concatenate[InlineKeyboardBuilder, P], object]
ReplyFn: TypeAlias = Callable[Concatenate[ReplyKeyboardBuilder, P], object]

InlineSyncFn: TypeAlias = Callable[P, InlineKeyboardMarkup]
ReplySyncFn: TypeAlias = Callable[P, ReplyKeyboardMarkup]


@overload
def dynamic_keyboard(
    *sizes: int,
    repeat: bool = True,
    adjust: bool = True,
    inline: Literal[True] = True,
    **builder_kwargs: object,
) -> Callable[[InlineFn[P]], InlineSyncFn[P]]: ...
@overload
def dynamic_keyboard(
    *sizes: int,
    repeat: bool = True,
    adjust: bool = True,
    inline: Literal[False],
    **builder_kwargs: object,
) -> Callable[[ReplyFn[P]], ReplySyncFn[P]]: ...
def dynamic_keyboard(  # type: ignore[misc]
    *sizes: int,
    repeat: bool = True,
    adjust: bool = True,
    inline: bool = True,
    **builder_kwargs: object,
) -> Callable[[Callable[Concatenate[BuilderT, P], object]], Callable[P, KeyboardMarkupT]]:
    """Decorator to build inline or reply keyboards via builder style.

    Example:

       .. code-block:: python

          @keyboard(inline=True)
          def markup(builder: InlineKeyboardBuilder, name: str | None = None):
              if name is not None:
                builder.button(text=f"Hello, {name}", callback_data="hello")
              builder.button(text="Back", callback_data="back")

    :param sizes: Row sizes passed to `adjust(...)`
    :param repeat: Whether adjust sizes should repeat
    :param adjust: Auto-adjust layout if True
    :param inline: If True, builds InlineKeyboardMarkup
    :param builder_kwargs: Extra args passed to `as_markup()`
    :returns: A wrapped function returning KeyboardMarkup
    """
    if not sizes:
        sizes = (1,)

    Builder = InlineKeyboardBuilder if inline else ReplyKeyboardBuilder

    def wrapper(fn: Callable[Concatenate[BuilderT, P], object]) -> Callable[P, KeyboardMarkupT]:
        @wraps(fn)
        def wrapped(*args: P.args, **kwargs: P.kwargs) -> KeyboardMarkupT:
            builder = Builder()
            fn(cast(BuilderT, builder), *args, **kwargs)
            if adjust:
                builder.adjust(*sizes, repeat=repeat)
            return builder.as_markup(**builder_kwargs)

        return wrapped

    return wrapper



from .data import RoleData
from .filter import RoleFilter
from .manager import RoleManager
from .providers import (
    BaseRoleProvider,
    IRoleProvider,
    MemoryRoleProvider,
    get_postgresql_provider,
    get_redis_provider,
    get_sqlite_provider,
)
from .roles import (
    ADMINISTRATOR,
    AVAILABLE_ROLES,
    AVAILABLE_ROLES_BY_SLUG,
    DEVELOPER,
    GUEST,
    MANAGER,
    MODERATOR,
    OWNER,
    SPONSOR,
    SUPPORT,
    TESTER,
)

__all__ = (
    "ADMINISTRATOR",
    "AVAILABLE_ROLES",
    "AVAILABLE_ROLES_BY_SLUG",
    "DEVELOPER",
    "GUEST",
    "MANAGER",
    "MODERATOR",
    "OWNER",
    "SPONSOR",
    "SUPPORT",
    "TESTER",
    "BaseRoleProvider",
    "IRoleProvider",
    "MemoryRoleProvider",
    "RoleData",
    "RoleFilter",
    "RoleManager",
    "get_postgresql_provider",
    "get_redis_provider",
    "get_sqlite_provider",
)



from collections.abc import Callable, Sequence
from functools import wraps
from typing import Literal, TypeAlias, cast, overload

from aiogram.types import (
    InlineKeyboardButton,
    InlineKeyboardMarkup,
    KeyboardButton,
    ReplyKeyboardMarkup,
)
from aiogram.utils.keyboard import (
    InlineKeyboardBuilder,
    ReplyKeyboardBuilder,
)
from typing_extensions import ParamSpec, TypeVar

__all__ = ("static_keyboard",)


P = ParamSpec("P")
R = TypeVar("R")

ButtonData: TypeAlias = str | tuple[str] | tuple[str, str]
LayoutRow: TypeAlias = Sequence[ButtonData]
LayoutReturn: TypeAlias = Sequence[ButtonData | LayoutRow]

InlineSyncFn: TypeAlias = Callable[P, InlineKeyboardMarkup]
ReplySyncFn: TypeAlias = Callable[P, ReplyKeyboardMarkup]

KeyboardMarkupT: TypeAlias = InlineKeyboardMarkup | ReplyKeyboardMarkup
BuilderFn: TypeAlias = Callable[P, KeyboardMarkupT]


def _get_button(data: ButtonData, *, inline: bool) -> InlineKeyboardButton | KeyboardButton:
    """Convert button data to the appropriate button type.

    :param data: Button representation (e.g. :code:`("Text", "callback_data")`, or :code:`"Text"` for ReplyKeyboardMarkup)
    :param inline: Whether to create InlineKeyboardButton or KeyboardButton
    :return: An instance of aiogram button
    :raises ValueError: if inline button data is invalid
    """
    if not inline:
        return KeyboardButton(text=data if isinstance(data, str) else data[0])

    if isinstance(data, str) or len(data) != 2:
        raise ValueError("InlineKeyboardButton must be tuple of (text, callback_data)")
    return InlineKeyboardButton(text=data[0], callback_data=data[1])


def _is_button(row: LayoutRow) -> bool:
    """Determine whether a given layout row represents a single button.

    A row is considered a single button if:
    - It is a `str` (used as the text for a reply button), or
    - It is a `tuple[str]` or `tuple[str, str]` (used for inline buttons with callback data)

    :param row: A row from the layout, which may be a button or a list of buttons
    :return: True if the row is a single button, False if it is a row of buttons
    """
    if (
        isinstance(row, tuple | list)
        and 0 < len(row) <= 2
        and all(isinstance(item, str) for item in row)
    ):
        return True
    return isinstance(row, str)


def _inject_layout(
    builder: InlineKeyboardBuilder | ReplyKeyboardBuilder,
    layout: LayoutReturn,
    *,
    inline: bool,
) -> None:
    """Add declarative layout to the keyboard builder.

    :param builder: The builder instance
    :param layout: List of button data or rows
    :param inline: Whether inline buttons are expected
    """
    expected_type = InlineKeyboardBuilder if inline else ReplyKeyboardBuilder
    if not isinstance(builder, expected_type):
        kind = "Inline" if inline else "Reply"
        raise ValueError(f"{kind} buttons are not supported for {type(builder).__name__}")

    for row in layout:
        if _is_button(row):
            button = _get_button(cast(ButtonData, row), inline=inline)
            builder.row(button, width=1)  # type: ignore[arg-type]
        else:
            buttons = [_get_button(data, inline=inline) for data in row]
            builder.row(*buttons, width=min(8, len(row)))  # type: ignore[arg-type]


@overload
def static_keyboard(
    inline: Literal[True] = True, **builder_kwargs: object
) -> Callable[[Callable[P, LayoutReturn]], InlineSyncFn[P]]: ...
@overload
def static_keyboard(
    inline: Literal[False], **builder_kwargs: object
) -> Callable[[Callable[P, LayoutReturn]], ReplySyncFn[P]]: ...
def static_keyboard(
    inline: bool = True, **builder_kwargs: object
) -> Callable[[Callable[P, LayoutReturn]], BuilderFn[P]]:
    """Decorator to build inline or reply keyboards via layout style.

    Example:

    .. code-block:: python

        @keyboard(inline=True)
        def markup():
            return [
                ("Top Button", "top_callback_data"),
                [("Left", "left"), ("Right", "right")],
            ]

    :param inline: If True, builds InlineKeyboardMarkup
    :param builder_kwargs: Extra args passed to `as_markup()`
    :returns: A wrapped function returning KeyboardMarkup
    """
    Builder = InlineKeyboardBuilder if inline else ReplyKeyboardBuilder

    def wrapper(fn: Callable[P, LayoutReturn]) -> BuilderFn[P]:
        @wraps(fn)
        def sync_fn(*args: P.args, **kwargs: P.kwargs) -> KeyboardMarkupT:
            value = fn(*args, **kwargs)

            if not isinstance(value, list):
                msg = "Function must return a list"
                raise ValueError(msg)

            builder = Builder()
            _inject_layout(builder, value, inline=inline)
            return builder.as_markup(**builder_kwargs)

        return sync_fn

    return wrapper



from collections.abc import Callable
from typing import NamedTuple, Optional

from pydantic import BaseModel

__all__ = (
    "NodeParts",
    "TreeChars",
    "TreeNode",
    "dot_paths_to_tree",
)


class TreeChars(BaseModel):
    """Tree branch characters for ASCII tree rendering."""

    last: str = "╚"
    middle: str = "╠"
    vertical: str = "║"


class NodeParts(NamedTuple):
    """Parts of a tree node."""

    icon: str
    folder: str
    prefix: str
    suffix: str


class TreeNode:
    """Tree node with support for prefixes, suffixes and customizable display."""

    def __init__(
        self,
        name: str,
        prefix: str = "",
        suffix: str = "",
        *,
        is_folder: bool = False,
    ) -> None:
        """Initialize TreeNode.

        :param name: Node name
        :type name: str
        :param prefix: Node prefix
        :type prefix: str
        :param suffix: Node suffix
        :type suffix: str
        :param is_folder: Whether the node represents a folder
        :type is_folder: bool
        """
        self.name = name
        self.prefix = prefix
        self.suffix = suffix
        self.is_folder = is_folder
        self.children: dict[str, TreeNode] = {}

    def add_child(
        self,
        name: str,
        prefix: str = "",
        suffix: str = "",
        *,
        is_folder: bool = False,
    ) -> "TreeNode":
        """Add a child node to this node.

        :param name: Child node name
        :type name: str
        :param prefix: Child node prefix
        :type prefix: str
        :param suffix: Child node suffix
        :type suffix: str
        :param is_folder: Whether the child node is a folder
        :type is_folder: bool
        :returns: The child node (new or existing)
        :rtype: TreeNode
        """
        if name not in self.children:
            self.children[name] = TreeNode(name, prefix, suffix, is_folder=is_folder)
        else:
            if prefix:
                self.children[name].prefix = prefix
            if suffix:
                self.children[name].suffix = suffix
            if is_folder:
                self.children[name].is_folder = True
        return self.children[name]

    def get_child(self, name: str) -> Optional["TreeNode"]:
        """Get a child node by name.

        :param name: Child node name
        :type name: str
        :returns: Child node if found, None otherwise
        :rtype: TreeNode | None
        """
        return self.children.get(name)

    def __repr__(self) -> str:
        """Return a string representation of the TreeNode.

        :returns: String representation of the TreeNode
        :rtype: str
        """
        return f"<TreeNode {self.name=} {self.prefix=} {self.suffix=} {self.is_folder=}>"


class AsciiTree:
    """ASCII tree renderer with customizable styling."""

    def __init__(
        self,
        folder_icon: str = "📁",
        folder_suffix: str = "/",
        tree_chars: TreeChars | None = None,
        *,
        sort: bool = True,
    ) -> None:
        """Initialize AsciiTree.

        :param folder_icon: Icon for folder nodes
        :type folder_icon: str
        :param folder_suffix: Suffix for folder nodes (e.g., "/")
        :type folder_suffix: str
        :param tree_chars: Tree characters (last, middle, vertical)
        :type tree_chars: TreeChars
        :param sort: Sort children alphabetically
        :type sort: bool
        """
        self.folder_icon = folder_icon
        self.folder_suffix = folder_suffix
        self.tree_chars = tree_chars or TreeChars()
        self.sort = sort

    def render(self, root: TreeNode) -> str:
        """Render the tree to ASCII format.

        :param root: Root node of the tree
        :type root: TreeNode
        :returns: ASCII rendered tree
        :rtype: str
        """
        if not root.children:
            return ""

        lines: list[str] = []
        self._render_children(root.children, "", lines)
        return "\n".join(lines)

    def _render_children(
        self,
        children: dict[str, TreeNode],
        prefix: str,
        lines: list[str],
    ) -> None:
        """Render child nodes recursively.

        :param children: Dictionary of child nodes
        :type children: dict[str, TreeNode]
        :param prefix: Current line prefix for indentation
        :type prefix: str
        :param lines: List to accumulate rendered lines
        :type lines: list[str]
        """
        nodes = list(children.items())
        if self.sort:
            nodes = sorted(nodes)
        last_index = len(nodes) - 1

        for i, (_, node) in enumerate(nodes):
            is_last = i == last_index
            branch = self.tree_chars.last if is_last else self.tree_chars.middle

            parts = NodeParts(
                icon=f"{self.folder_icon} " if node.is_folder and self.folder_icon else "",
                folder=self.folder_suffix if node.is_folder else "",
                prefix=f"{node.prefix} " if node.prefix and not node.is_folder else "",
                suffix=f" {node.suffix}" if node.suffix and not node.is_folder else "",
            )

            line = f"{prefix}{branch} {parts.prefix}{parts.icon}{node.name}{parts.folder}{parts.suffix}"
            lines.append(line)

            if node.children:
                empty = " " * 4
                extension = empty if is_last else f"{self.tree_chars.vertical}" + empty[1:]
                self._render_children(node.children, prefix + extension, lines)


def dot_paths_to_tree(
    paths: list[str] | dict[str, str],
    prefix_callback: Callable[[str], str | None] | None = None,
    suffix_callback: Callable[[str], str | None] | None = None,
) -> TreeNode:
    """Convert dot-notation paths to a tree structure.

    :param paths: List of paths ["folder.child", "folder.folder.child"]
    :type paths: list[str]
    :param prefix_callback: Function to get node prefixes
    :type prefix_callback: Callable[[str], str | None] | None
    :param suffix_callback: Function to get node suffixes
    :type suffix_callback: Callable[[str], str | None] | None
    :returns: Root node of the tree
    :rtype: TreeNode
    """
    root = TreeNode("root", is_folder=True)

    for path in paths:
        parts = path.split(".")
        current_node = root

        for i, part in enumerate(parts):
            is_last_part = i == len(parts) - 1
            current_path = ".".join(parts[: i + 1])

            prefix, suffix = "", ""
            if prefix_callback:
                prefix = prefix_callback(current_path) or ""
            if suffix_callback:
                suffix = suffix_callback(current_path) or ""

            is_folder = not is_last_part or any(
                other_path.startswith(current_path + ".")
                for other_path in paths
                if other_path != current_path
            )

            current_node = current_node.add_child(part, prefix, suffix, is_folder=is_folder)

    return root



from __future__ import annotations

from typing import TYPE_CHECKING

from aiogram.client.default import Default
from aiogram.types import InputMediaPhoto

from raito.plugins.pagination.enums import PaginationMode

from .base import BasePaginator

if TYPE_CHECKING:
    from aiogram.types import (
        InlineKeyboardMarkup,
        InputFileUnion,
        Message,
        MessageEntity,
        ReplyParameters,
    )


__all__ = ("PhotoPaginator",)


class PhotoPaginator(BasePaginator):
    """Photo paginator."""

    def _validate_parameters(
        self,
        name: str,
        current_page: int,
        total_pages: int | None,
        limit: int,
    ) -> None:
        """Validate paginator parameters.

        :param name: pagination name
        :type name: str
        :param current_page: current page number
        :type current_page: int
        :param total_pages: total pages count
        :type total_pages: int | None
        :param limit: items per page
        :type limit: int
        :raises ValueError: if parameters are invalid
        """
        if limit > 1024:
            raise ValueError("limit must be less than or equal to 1024")

        return super()._validate_parameters(
            name=name,
            current_page=current_page,
            total_pages=total_pages,
            limit=limit,
        )

    @property
    def mode(self) -> PaginationMode:
        """Get text pagination mode.

        :return: pagination mode
        :rtype: PaginationMode
        """
        return PaginationMode.PHOTO

    async def answer(
        self,
        photo: InputFileUnion,
        caption: str | None = None,
        parse_mode: str | Default | None = None,
        caption_entities: list[MessageEntity] | None = None,
        show_caption_above_media: bool | Default | None = None,
        has_spoiler: bool | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | Default | None = None,
        allow_paid_broadcast: bool | None = None,
        message_effect_id: str | None = None,
        reply_parameters: ReplyParameters | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
        allow_sending_without_reply: bool | None = None,
        reply_to_message_id: int | None = None,
    ) -> Message:
        """Send or edit paginated message.

        :param photo: photo file
        :type photo: InputFileUnion
        :param caption: message text
        :type caption: str
        :param parse_mode: text parse mode
        :type parse_mode: str | Default | None
        :param caption_entities: message entities
        :type caption_entities: list[MessageEntity] | None
        :param show_caption_above_media: show caption above media
        :type show_caption_above_media: bool | None
        :param has_spoiler: has spoiler
        :type has_spoiler: bool | None
        :param disable_notification: disable notification
        :type disable_notification: bool | None
        :param protect_content: protect content
        :type protect_content: bool | Default | None
        :param allow_paid_broadcast: allow paid broadcast
        :type allow_paid_broadcast: bool | None
        :param message_effect_id: message effect id
        :type message_effect_id: int | None
        :param reply_parameters: reply parameters
        :type reply_parameters: ReplyParameters | None
        :param reply_markup: custom reply markup
        :type reply_markup: InlineKeyboardMarkup | None
        :param allow_sending_without_reply: allow sending without reply
        :type allow_sending_without_reply: bool | None
        :param reply_to_message_id: reply to message id
        :type reply_to_message_id: int | None
        :return: paginated message
        :rtype: Message
        :raises RuntimeError: if bot instance not set
        """
        if not self.bot:
            raise RuntimeError("Bot not set via PaginatorMiddleware")

        parse_mode = parse_mode or Default("parse_mode")
        show_caption_above_media = show_caption_above_media or Default("show_caption_above_media")
        protect_content = protect_content or Default("protect_content")

        reply_markup = reply_markup or self.build_navigation()

        if self.existing_message is None:
            self.existing_message = await self.bot.send_photo(
                chat_id=self.chat_id,
                photo=photo,
                caption=caption,
                parse_mode=parse_mode,
                caption_entities=caption_entities,
                show_caption_above_media=show_caption_above_media,
                has_spoiler=has_spoiler,
                disable_notification=disable_notification,
                protect_content=protect_content,
                allow_paid_broadcast=allow_paid_broadcast,
                message_effect_id=message_effect_id,
                reply_parameters=reply_parameters,
                reply_markup=reply_markup,
                allow_sending_without_reply=allow_sending_without_reply,
                reply_to_message_id=reply_to_message_id,
            )
        else:
            await self.existing_message.edit_media(
                media=InputMediaPhoto(media=photo, caption=caption, reply_markup=reply_markup),
                caption=caption,
                reply_markup=reply_markup,
            )

        return self.existing_message



from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, Any, TypeAlias, TypeVar

from aiogram.dispatcher.event.bases import REJECTED
from aiogram.dispatcher.event.handler import HandlerObject
from aiogram.dispatcher.middlewares.base import BaseMiddleware
from aiogram.filters.command import CommandObject
from aiogram.types import Message, TelegramObject

from raito.utils.helpers.command_help import get_command_help

if TYPE_CHECKING:
    from raito.core.raito import Raito

DataT = TypeVar("DataT", bound=dict[str, Any])
R = TypeVar("R")
ParamT: TypeAlias = type[int] | type[str] | type[bool] | type[float]


__all__ = ("CommandMiddleware",)


class CommandMiddleware(BaseMiddleware):
    """Middleware for command-related features.

    - Supports automatic parameter parsing from text based on the :code:`raito__params` flag.

    *Can be extended with additional logic in the future*
    """

    def __init__(self) -> None:
        """Initialize CommandMiddleware."""

    def _unpack_params(
        self,
        command: CommandObject,
        params: dict[str, ParamT],
        data: DataT,
    ) -> DataT:
        """Unpack command parameters into the metadata.

        :param handler_object: Handler object
        :param event: Telegram message
        :param data: Current metadata
        :return: Updated context with parsed parameters
        :raises ValueError, IndexError: If parameter is missing or invalid
        """
        args = command.args.split() if command.args else []
        for i, (key, value_type) in enumerate(params.items()):
            arg = args[i]
            if value_type is bool:
                bool_value: bool = arg.lower() in ("true", "yes", "on", "1", "ok", "+")
                data[key] = bool_value
            else:
                data[key] = value_type(arg)

        return data

    async def _send_help_message(
        self,
        handler_object: HandlerObject,
        command: CommandObject,
        params: dict[str, ParamT],
        event: Message,
        data: dict[str, Any],
    ) -> None:
        description = handler_object.flags.get("raito__description")
        raito: Raito | None = data.get("raito")

        if raito is not None and raito.command_parameters_error.handlers:
            target = {"handler": handler_object, "command": command}
            await raito.command_parameters_error.trigger(event, target=target)
        else:
            await event.reply(
                get_command_help(command, params, description=description),
                parse_mode="HTML",
            )

    async def __call__(
        self,
        handler: Callable[[TelegramObject, dict[str, Any]], Awaitable[R]],
        event: TelegramObject,
        data: dict[str, Any],
    ) -> R | None:
        """Process incoming events with command logic.

        :param handler: Next handler in the middleware chain
        :type handler: Callable
        :param event: Telegram event (Message or CallbackQuery)
        :type event: TelegramObject
        :param data: Additional data passed through the middleware chain
        :type data: dict[str, Any]
        :return: Handler result if not throttled, None if throttled
        """
        if not isinstance(event, Message):
            return await handler(event, data)

        handler_object: HandlerObject | None = data.get("handler")
        if handler_object is None:
            return await handler(event, data)

        command: CommandObject | None = data.get("command")
        if command is None:
            return await handler(event, data)

        params: dict[str, ParamT] | None = handler_object.flags.get("raito__params")
        if params:
            try:
                data = self._unpack_params(command, params, data)
            except (ValueError, IndexError):
                await self._send_help_message(handler_object, command, params, event, data)
                return REJECTED

        return await handler(event, data)



from __future__ import annotations

from collections import defaultdict
from contextlib import suppress
from typing import TYPE_CHECKING, NamedTuple

from aiogram import Bot
from aiogram.dispatcher.event.handler import HandlerObject
from aiogram.types import (
    BotCommand,
    BotCommandScopeChat,
    BotCommandScopeDefault,
    BotCommandScopeUnion,
)
from aiogram.utils.i18n.context import gettext

from raito.plugins.roles.data import RoleData
from raito.utils import loggers

if TYPE_CHECKING:
    from aiogram.filters.command import Command
    from aiogram.utils.i18n.lazy_proxy import LazyProxy  # type: ignore

    from raito.plugins.roles.manager import RoleManager


__all__ = ("register_bot_commands",)


class _CommandMeta(NamedTuple):
    """Command metadata.

    :param command: Command name (without prefix)
    :type command: str
    :param description: Localized or plain description
    :type description: str
    :param role: Minimum required role (if any)
    :type role: Role | None
    """

    command: str
    description: str
    roles: list[RoleData]


def _extract_command_metadata(handler: HandlerObject) -> _CommandMeta | None:
    """Extract command metadata from a handler's flags.

    :param handler: Message handler object
    :type handler: HandlerObject
    :return: CommandMeta instance or None
    :rtype: CommandMeta | None
    """
    commands: list[Command] | None = handler.flags.get("commands")
    if not commands or not commands[0].commands:
        return None

    if handler.flags.get("raito__hidden"):
        return None

    description: LazyProxy | str | None = handler.flags.get("raito__description", "—")
    roles: list[RoleData] | None = handler.flags.get("raito__roles")

    return _CommandMeta(
        command=commands[0].commands[0],
        description=str(description).strip(),
        roles=roles or [],
    )


def _format_description(meta: _CommandMeta, text: str) -> str:
    """Format description with role emoji if available.

    :param meta: CommandMeta instance
    :type meta: CommandMeta
    :param text: Description string
    :type text: str
    :return: Formatted description
    :rtype: str
    """
    if meta.roles is None:
        return text

    emojis = "".join([r.emoji for r in meta.roles])
    return f"[{emojis}] {text}"


async def _apply_bot_commands(
    bot: Bot,
    meta_entries: list[_CommandMeta],
    scope: BotCommandScopeUnion,
    locale: str | None = None,
) -> None:
    """Set commands for a given scope and locale.

    :param bot: Bot instance
    :type bot: Bot
    :param meta_entries: List of command metadata
    :type meta_entries: list[_CommandMeta]
    :param locale: Locale string (e.g., "en", "ru")
    :type locale: str
    :param scope: Scope for which to set commands
    :type scope: BotCommandScopeUnion
    """
    loggers.commands.debug(
        "Applying %s commands for user %s with locale %s",
        len(meta_entries),
        getattr(scope, "chat_id", None),
        locale,
    )

    bot_commands: list[BotCommand] = []
    for meta in meta_entries:
        description = meta.description
        with suppress(LookupError):
            description = gettext(description, locale=locale)

        bot_commands.append(
            BotCommand(
                command=meta.command,
                description=_format_description(meta, description),
            )
        )

    await bot.set_my_commands(commands=bot_commands, scope=scope, language_code=locale)


async def register_bot_commands(
    role_manager: RoleManager,
    bot: Bot,
    handlers: list[HandlerObject],
    locales: list[str],
) -> None:
    """Register localized bot commands across roles and user scopes.

    :param role_manager: RoleManager instance
    :type role_manager: RoleManager
    :param bot: Aiogram Bot instance
    :type bot: Bot
    :param handlers: List of message handler objects
    :type handlers: list[HandlerObject]
    :param locales: List of supported locales (e.g., "en", "ru")
    :type locales: list[str]
    """
    role_commands: dict[RoleData | None, list[_CommandMeta]] = defaultdict(list)

    for handler in handlers:
        meta = _extract_command_metadata(handler)
        if not meta:
            continue

        if not meta.roles:
            role_commands[None].append(meta)

        for role in meta.roles:
            role_commands[role].append(meta)

    role_users: dict[RoleData, set[int]] = defaultdict(set)
    for role in role_commands:  # type: ignore
        if role is None:
            continue

        users = await role_manager.get_users(bot.id, role.slug)
        role_users[role].update(users)

    for locale in locales:
        for role, users in role_users.items():
            commands = role_commands.get(role, [])
            for user_id in users:
                await _apply_bot_commands(
                    bot=bot,
                    meta_entries=commands,
                    scope=BotCommandScopeChat(chat_id=user_id),
                    locale=locale,
                )

        await _apply_bot_commands(
            bot=bot,
            meta_entries=role_commands.get(None, []),
            scope=BotCommandScopeDefault(),
            locale=locale,
        )

    await _apply_bot_commands(
        bot=bot,
        meta_entries=role_commands.get(None, []),
        scope=BotCommandScopeDefault(),
        locale=None,
    )



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from watchfiles import Change, awatch

from raito.utils import loggers

from .loader import RouterLoader
from .parser import RouterParser

if TYPE_CHECKING:
    from collections.abc import Generator

    from aiogram import Dispatcher

    from raito.utils.types import StrOrPath

__all__ = ("RouterManager",)


class RouterManager:
    """Manages multiple routers and file watching."""

    def __init__(self, dispatcher: Dispatcher) -> None:
        """Initialize the RouterManager.

        :param dispatcher: Aiogram dispatcher instance
        :type dispatcher: Dispatcher
        """
        self.dispatcher = dispatcher

        self.loaders: dict[str, RouterLoader] = {}

    def resolve_paths(self, directory: StrOrPath) -> Generator[StrOrPath]:  # pyright: ignore [reportInvalidTypeArguments]
        """Recursively resolve all router paths in a directory.

        Scans the given directory recursively for Python files that can contain routers.
        Ignores files and directories starting with underscore `_`

        :param directory: Directory to scan for router files
        :type directory: StrOrPath
        :yield: Path objects for router files found in the directory
        :rtype: Generator[StrOrPath, None, None]
        """
        dir_path = Path(directory)

        for item in dir_path.iterdir():
            if item.name.startswith("_"):  # ignore files with prefix _
                continue

            if item.is_file() and item.suffix == ".py":
                yield item
            elif item.is_dir():
                yield from self.resolve_paths(item)

    async def load_routers(self, directory: StrOrPath) -> None:
        """Load all routers from a directory.

        Scans the directory for Python files containing routers, extracts them,
        handles name conflicts by adding unique suffixes, and registers them
        with the dispatcher.

        :param directory: Directory containing router files
        :type directory: StrOrPath
        :raises AttributeError: If a router doesn't have a name attribute
        """
        dir_path = Path(directory)

        for file_path in self.resolve_paths(dir_path):
            try:
                router = RouterParser.extract_router(file_path)
            except (ModuleNotFoundError, TypeError) as exc:
                loggers.routers.error(
                    "Error while trying to load router from %s: %s", file_path, exc
                )
                continue

            try:
                unique_name: str = router.name
            except AttributeError as e:
                msg = "The router has no name"
                raise AttributeError(msg) from e

            if unique_name in self.loaders:
                suffix = hex(id(router))
                unique_name = f"{router.name}_{suffix}"
                loggers.routers.warning(
                    "Duplicate router name: %s. Will rename to %s...",
                    router.name,
                    unique_name,
                )
                router.name = unique_name

            loader = RouterLoader(
                unique_name,
                file_path,
                self.dispatcher,
                router=router,
            )
            loader.load()
            self.loaders[unique_name] = loader
            loggers.routers.debug("Router loaded: %s", unique_name)

    async def start_watchdog(self, directory: StrOrPath) -> None:
        """Start file watching service.

        Monitors the specified directory for file changes and automatically
        reloads routers when their corresponding files are modified.

        :param directory: Directory to watch for changes
        :type directory: StrOrPath
        """
        loggers.routers.info("Router watchdog started for: %s", directory)
        async for changes in awatch(directory, step=500):
            for event_type, changed_path in changes:
                path_object = Path(changed_path).resolve()

                current_loader: RouterLoader | None = None
                for loader in self.loaders.values():
                    if Path(loader.path).resolve() == path_object:
                        current_loader = loader
                        break

                if not current_loader:
                    loggers.routers.debug(
                        "File changed: %s. No routers found.",
                        changed_path,
                    )
                    continue

                if event_type in (Change.modified, Change.added):
                    loggers.routers.debug(
                        "File changed: %s. Reloading router '%s'...",
                        changed_path,
                        current_loader.name,
                    )

                    try:
                        await current_loader.reload()
                    except Exception as exc:  # noqa: BLE001
                        loggers.routers.error(
                            "Router '%s' has an error '%s'. Skipping...",
                            current_loader.path,
                            exc,
                        )
                        continue

                elif event_type == Change.deleted:
                    loggers.routers.debug(
                        "File removed: %s. Unloading router '%s'...",
                        changed_path,
                        current_loader.name,
                    )
                    current_loader.unload()
                break



from __future__ import annotations

from raito.plugins.roles.data import RoleData

from .providers.protocol import IRoleProvider
from .roles import AVAILABLE_ROLES, AVAILABLE_ROLES_BY_SLUG

__all__ = ("RoleManager",)


class RoleManager:
    """Central manager for role-based access control in Raito."""

    def __init__(self, provider: IRoleProvider, developers: list[int] | None = None) -> None:
        """Initialize RoleManager.

        :param provider: Role provider instance for persistent storage
        :type provider: IRoleProvider
        """
        self.provider = provider
        self.developers = developers or []

    async def migrate(self) -> None:
        """Run provider migrations."""
        await self.provider.migrate()

    async def get_role(self, bot_id: int, user_id: int) -> str | None:
        """Get the role for a specific user.

        :param bot_id: The Telegram bot ID
        :param user_id: The Telegram user ID
        :return: The role slug or None if not found
        """
        return await self.provider.get_role(bot_id, user_id)

    async def can_manage_roles(self, bot_id: int, user_id: int) -> bool:
        """Check whether the user can manage other users' roles.

        :param bot_id: The Telegram bot ID
        :param user_id: The Telegram user ID
        :return: True if user can manage roles, False otherwise
        """
        return await self.has_any_roles(
            bot_id,
            user_id,
            "developer",
            "administrator",
            "owner",
        )

    async def assign_role(
        self,
        bot_id: int,
        initiator_id: int,
        target_id: int,
        role_slug: str,
    ) -> None:
        """Assign a role to a user.

        :param bot_id: The Telegram bot ID
        :param initiator_id: The Telegram user ID of the initiator
        :param target_id: The Telegram user ID of the target
        :param role_slug: The role to assign
        :raises PermissionError: If the user does not have permission to assign role
        """
        if initiator_id == target_id:
            msg = "You cannot assign your own role."
            raise PermissionError(msg)

        if not await self.can_manage_roles(bot_id, initiator_id):
            msg = "You do not have permission to assign roles."
            raise PermissionError(msg)

        await self.provider.set_role(bot_id, target_id, role_slug)

    async def revoke_role(self, bot_id: int, initiator_id: int, target_id: int) -> None:
        """Revoke a user's role.

        :param bot_id: The Telegram bot ID
        :param initiator_id: The Telegram user ID of the initiator
        :param target_id: The Telegram user ID of the target
        :raises PermissionError: If the user does not have permission to revoke roles
        """
        if initiator_id == target_id:
            msg = "You cannot revoke your own role."
            raise PermissionError(msg)

        if not await self.can_manage_roles(bot_id, initiator_id):
            msg = "You do not have permission to assign roles."
            raise PermissionError(msg)

        await self.provider.remove_role(bot_id, target_id)

    async def has_any_roles(self, bot_id: int, user_id: int, *roles: str) -> bool:
        """Check if a user has any of the specified roles.

        :param bot_id: The Telegram bot ID
        :param user_id: The Telegram user ID
        :param roles: Roles to check for
        :returns: True if user has any of the roles, False otherwise
        """
        if user_id in self.developers and "developer" in roles:
            return True

        role = await self.get_role(bot_id, user_id)
        return role in roles

    async def has_role(self, bot_id: int, user_id: int, role_slug: str) -> bool:
        """Check if a user has the specified roles.

        :param bot_id: The Telegram bot ID
        :param user_id: The Telegram user ID
        :param role_slug: Role to check for
        :returns: True if user has the role, False otherwise
        """
        if user_id in self.developers and role_slug == "developer":
            return True

        slug = await self.get_role(bot_id, user_id)
        return slug == role_slug

    async def get_users(self, bot_id: int, role: str) -> set[int]:
        """Get a list of users with a specific role.

        :param bot_id: The Telegram bot ID
        :param role: The role to check for
        :returns: A list of Telegram user IDs
        """
        users = await self.provider.get_users(bot_id, role)
        unique_users = set(users)

        if role == "developer":
            unique_users.update(self.developers)

        return unique_users

    @property
    def available_roles(self) -> list[RoleData]:
        """Get a list of available roles.

        :returns: A list of roles
        """
        return AVAILABLE_ROLES

    def get_role_data(self, slug: str) -> RoleData:
        """Get data of specified role.

        :returns: A data of role
        :raises: ...
        """
        return AVAILABLE_ROLES_BY_SLUG[slug]



from __future__ import annotations

from asyncio import sleep
from typing import TYPE_CHECKING

from aiogram import Router, html

from raito.plugins.commands import description, hidden, params
from raito.plugins.roles.roles import DEVELOPER
from raito.utils.filters import RaitoCommand

if TYPE_CHECKING:
    from aiogram.types import Message

    from raito.core.raito import Raito

router = Router(name="raito.management.load")


@router.message(RaitoCommand("load"), DEVELOPER)
@description("Loads a router by name")
@params(name=str)
@hidden
async def load_router(message: Message, raito: Raito, name: str) -> None:
    router_loader = raito.router_manager.loaders.get(name)
    if not router_loader:
        await message.answer(f"🔎 Router {html.bold(name)} not found", parse_mode="HTML")
        return

    msg = await message.answer(f"📦 Loading router {html.bold(name)}...", parse_mode="HTML")
    router_loader.load()
    await sleep(0.5)
    await msg.edit_text(f"✅ Router {html.bold(name)} loaded", parse_mode="HTML")



from .data import PaginationCallbackData
from .decorator import on_pagination
from .enums import PaginationMode
from .middleware import PaginatorMiddleware
from .paginators import BasePaginator, InlinePaginator, PhotoPaginator, TextPaginator
from .util import get_paginator

__all__ = (
    "BasePaginator",
    "InlinePaginator",
    "PaginationCallbackData",
    "PaginationMode",
    "PaginatorMiddleware",
    "PhotoPaginator",
    "TextPaginator",
    "get_paginator",
    "on_pagination",
)



from __future__ import annotations

from typing import TYPE_CHECKING

from aiogram import F, Router

from .data import PaginationCallbackData

if TYPE_CHECKING:
    from aiogram.dispatcher.event.handler import CallbackType

__all__ = ("on_pagination",)


def on_pagination(router: Router, name: str, *filters: CallbackType) -> CallbackType:
    """Register pagination handler for specific name.

    :param router: aiogram router
    :type router: Router
    :param name: pagination name
    :type name: str
    :return: decorator function
    :rtype: CallbackType
    """
    return router.callback_query(
        PaginationCallbackData.filter(F.name == name),
        *filters,
        flags={"raito__is_pagination": True},
    )



from .base_router import BaseRouter
from .loader import RouterLoader
from .manager import RouterManager
from .parser import RouterParser

__all__ = (
    "BaseRouter",
    "RouterLoader",
    "RouterManager",
    "RouterParser",
)



from .enums import PaginationMode
from .paginators import BasePaginator, InlinePaginator, PhotoPaginator, TextPaginator

__all__ = ("get_paginator",)


def get_paginator(mode: PaginationMode) -> type[BasePaginator]:
    """Get paginator class by mode.

    :param mode: pagination mode
    :type mode: PaginationMode
    :return: paginator class
    :rtype: type[BasePaginator]
    :raises ValueError: if mode is not supported
    """
    if mode == PaginationMode.INLINE:
        return InlinePaginator
    if mode == PaginationMode.TEXT:
        return TextPaginator
    if mode == PaginationMode.PHOTO:
        return PhotoPaginator

    raise ValueError(f"Unsupported pagination mode: {mode}")



from __future__ import annotations

from aiogram.dispatcher.flags import Flag, FlagDecorator

from .middleware import THROTTLING_MODE

__all__ = ("limiter",)


def limiter(rate_limit: float, mode: THROTTLING_MODE = "user") -> FlagDecorator:
    """Attach a rate limit to the command handler.

    This decorator sets a custom rate limit (in seconds) for a specific command handler.

    :param rate_limit: Minimum delay between invokes (in seconds)
    :param mode: Throttling key type: 'user', 'chat', or 'bot'
    :return: Combined FlagDecorator
    """
    data = {"rate_limit": rate_limit, "mode": mode}
    return FlagDecorator(Flag("raito__limiter", value=data))



from .flags import description, hidden, params
from .middleware import CommandMiddleware
from .registration import register_bot_commands

__all__ = (
    "CommandMiddleware",
    "description",
    "hidden",
    "params",
    "register_bot_commands",
)
